import os
import pathlib
from typing import List, Optional, Union

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.colors import ListedColormap, Normalize
from matplotlib.figure import Figure
from tqdm.auto import tqdm

from toffy import settings
from toffy.qc_comp import QCTMA, QCControlMetrics


def visualize_qc_metrics(
    metric_name: str,
    qc_metric_dir: Union[str, pathlib.Path],
    save_dir: Union[str, pathlib.Path],
    channel_filters: Optional[List[str]] = ["chan_"],
    axes_font_size: int = 16,
    wrap: int = 6,
    dpi: int = 300,
    return_plot: bool = False,
) -> Optional[sns.FacetGrid]:
    """Visualize the barplot of a specific QC metric.

    Args:
        metric_name (str):
            The name of the QC metric to plot. Used as the y-axis label. Options include:
            `"Non-zero mean intensity"`, `"Total intensity"`, `"99.9% intensity value"`.
        qc_metric_dir (Union[str, pathlib.Path]):
            The path to the directory containing the `'combined_{qc_metric}.csv'` files
        save_dir (Optional[Union[str, pathlib.Path]], optional):
            The name of the directory to save the plot to. Defaults to None.
        channel_filters (List[str], optional):
            A list of channels to filter out.
        axes_font_size (int, optional):
            The font size of the axes labels. Defaults to 16.
        wrap (int, optional):
            The number of plots to display per row. Defaults to 6.
        dpi (Optional[int], optional):
            The resolution of the image to use for saving. Defaults to None.
        return_plot (bool):
            If `True`, this will return the plot. Defaults to `False`

    Raises:
        ValueError:
            When an invalid metric is provided.
        FileNotFoundError:
            The QC metric directory `qc_metric_dir` does not exist.
        FileNotFoundError:
            The QC metric `combined_csv` file is does not exist in `qc_metric_dir`.

    Returns:
        Optional[sns.FacetGrid]: Returns the Seaborn FacetGrid catplot of the QC metrics.
    """
    # verify the metric provided is valid
    if metric_name not in settings.QC_COLUMNS:
        raise ValueError(
            "Invalid metric %s provided, must be set to 'Non-zero mean intensity', "
            "'Total intensity', or '99.9%% intensity value'" % metric_name
        )

    # verify the path to the QC metric datasets exist
    if not os.path.exists(qc_metric_dir):
        raise FileNotFoundError("qc_metric_dir %s does not exist" % qc_metric_dir)

    # get the file name of the combined QC metric .csv file to use
    qc_metric_index = settings.QC_COLUMNS.index(metric_name)
    qc_metric_suffix = settings.QC_SUFFIXES[qc_metric_index]
    qc_metric_path = os.path.join(qc_metric_dir, "combined_%s.csv" % qc_metric_suffix)

    # ensure the user set the right qc_metric_dir
    if not os.path.exists(qc_metric_path):
        raise FileNotFoundError(
            "Could not locate %s, ensure qc_metric_dir is correct" % qc_metric_path
        )

    # read in the QC metric data
    qc_metric_df = pd.read_csv(qc_metric_path)

    # filter out naturally-occurring elements as well as Noodle
    qc_metric_df = qc_metric_df[~qc_metric_df["channel"].isin(settings.QC_CHANNEL_IGNORE)]

    # filter out any channel in the channel_filters list
    if channel_filters is not None:
        qc_metric_df: pd.DataFrame = qc_metric_df[
            ~qc_metric_df["channel"].str.contains("|".join(channel_filters))
        ]

    # catplot allows for easy facets on a barplot
    qc_fg: sns.FacetGrid = sns.catplot(
        x="fov",
        y=metric_name,
        col="channel",
        col_wrap=wrap,
        data=qc_metric_df,
        kind="bar",
        color="black",
        sharex=True,
        sharey=False,
    )

    # remove the 'channel =' in each subplot title
    qc_fg.set_titles(template="{col_name}")
    qc_fg.figure.supxlabel(t="fov", x=0.5, y=0, ha="center", size=axes_font_size)
    qc_fg.figure.supylabel(t=f"{metric_name}", x=0, y=0.5, va="center", size=axes_font_size)
    qc_fg.set(xticks=[])

    # per Erin's visualization remove the default axis title on the y-axis
    # and instead show 'fov' along x-axis and the metric name along the y-axis (overarching)
    qc_fg.set_axis_labels(x_var="", y_var="")
    qc_fg.set_xticklabels([])

    # save the figure always
    # Return the figure if specified.
    qc_fg.savefig(os.path.join(save_dir, f"{metric_name}_barplot_stats.png"), dpi=dpi)

    if return_plot:
        return qc_fg


def qc_tmas_metrics_plot(
    qc_tmas: QCTMA,
    tmas: List[str],
    save_figure: bool = False,
    dpi: int = 300,
) -> None:
    """Produces the QC TMA metrics plot for a given set of QC metrics applied to a user specified
    TMA. The figures are saved in `qc_tma_metrics_dir/figures`.

    Args:
        qc_tmas (QCTMA): The class which contains the QC TMA data, filepaths, and methods.
        QC matrix.
        tmas (str): The TMAs to plot the QC metrics for.
        save_figure (bool, optional): If `True`, the figure is saved in a subdirectory in the
        `QCTMA.qc_tma_metrics_dir` directory. Defaults to `False`.
        dpi (int, optional): Dots per inch, the resolution of the image. Defaults to 300.
    """
    if save_figure:
        fig_dir: pathlib.Path = pathlib.Path(qc_tmas.metrics_dir) / "figures"
        fig_dir.mkdir(parents=True, exist_ok=True)

    # also plot averages
    tmas.append("cross_TMA_averages")

    with tqdm(total=len(tmas), desc="Plotting QC TMA Metric Z-scores", unit="TMAs") as pbar:
        for tma in tmas:
            _qc_tma_metrics_plot(qc_tmas, tma, fig_dir=fig_dir, save_figure=save_figure, dpi=dpi)
            pbar.set_postfix(TMA=tma)
            pbar.update(n=1)


def _qc_tma_metrics_plot(
    qc_tmas: QCTMA,
    tma: str,
    fig_dir: pathlib.Path,
    save_figure: bool = False,
    dpi: int = 300,
) -> None:
    """Produces the QC TMA metrics plot for a given set of QC metrics applied to a user specified
    TMA. The figures are saved in `qc_tma_metrics_dir/figures`.

    Args:
        qc_tmas (QCTMA): The class which contains the QC TMA data, filepaths, and methods.
        tma (str): The TMA to plot the metrics for.
        fig_dir (pathlib.Path) : Path of where to save the plots.
        save_figure (bool, optional): If `True`, the figure is saved in a subdirectory in the
        `qc_tma_metrics_dir` directory. Defaults to `False`.
        dpi (int, optional): Dots per inch, the resolution of the image. Defaults to 300.
    """
    for qc_metric, suffix in zip(qc_tmas.qc_cols, qc_tmas.qc_suffixes):
        qc_tma_data: np.ndarray = qc_tmas.tma_avg_zscores[tma].loc[qc_metric].values

        # Set up the Figure for multiple axes
        fig: Figure = plt.figure(dpi=dpi)
        qc_tma_data = qc_tma_data.round(decimals=2)

        # Heatmap
        _norm = Normalize(vmin=-1, vmax=1)
        ax_heatmap = fig.add_subplot()
        sns.heatmap(
            data=qc_tma_data,
            square=True,
            ax=ax_heatmap,
            linewidths=1,
            linecolor="black",
            annot_kws={"size": 7},
            cbar_kws={"shrink": 0.5},
            annot=True,
            cmap=sns.color_palette(palette="vlag", as_cmap=True),
            norm=_norm,
        )
        # Set ticks
        ax_heatmap.set_xticks(
            ticks=ax_heatmap.get_xticks(),
            labels=[f"{i+1}" for i in range(qc_tma_data.shape[1])],
            rotation=0,
        )

        ax_heatmap.set_yticks(
            ticks=ax_heatmap.get_yticks(),
            labels=[f"{i+1}" for i in range(qc_tma_data.shape[0])],
            rotation=0,
        )

        ax_heatmap.set_xlabel("Column")
        ax_heatmap.set_ylabel("Row")
        if tma == "cross_TMA_averages":
            ax_heatmap.set_title("Average Scores Across All TMAs\n")
        else:
            ax_heatmap.set_title(f"{tma} - Average {qc_metric}\n")

        if save_figure:
            fig.savefig(
                fname=fig_dir / f"{tma}_{suffix}.png",
                dpi=dpi,
                bbox_inches="tight",
            )
            plt.close(fig)


def longitudinal_control_heatmap(
    qc_control: QCControlMetrics,
    control_sample_name: str,
    save_lc_df: bool = False,
    save_figure: bool = False,
    display_figure: bool = False,
    figsize: tuple[int, int] = (12, 12),
    dpi: int = 300,
) -> None:
    """Generates a heatmap of the QC metrics for the QC Control FOVs.

    Args:
        qc_control (QCControlMetrics): The class which contains the QC LC data, filepaths
        , and methods.
        control_sample_name (List[str]): A list of tissues to plot the QC metrics for.
        save_lc_df: (bool, optional): If `True`, the longitudinal control data is saved as a `csv` in
        `qc_control.metrics_dir`. Defaults to `False`.
        save_figure (bool, optional): If `True`, the figure is saved in a subdirectory in the
        `longitudinal_control_metrics_dir` directory. Defaults to `False`.
        display_figure (bool, optional): If `True`, the figure is displayed. Defaults to `False`.
        figsize: (tuple[int, int], optional): The size of the figure. Defaults to (12, 12).
        dpi (int, optional): Dots per inch, the resolution of the image. Defaults to 300.

    Raises:
        ValueError: Raised when the input tissues are not a list of strings.
    """
    if control_sample_name is None or not isinstance(control_sample_name, str):
        raise ValueError("The control sample name must be string.")
    if save_figure:
        fig_dir: pathlib.Path = pathlib.Path(qc_control.metrics_dir) / "figures"
        fig_dir.mkdir(parents=True, exist_ok=True)

    for qc_col, qc_suffix in zip(qc_control.qc_cols, qc_control.qc_suffixes):
        # Try to read the transformed df if it exists
        try:
            t_df: pd.DataFrame = pd.read_csv(
                os.path.join(
                    qc_control.metrics_dir, f"{control_sample_name}_transformed_{qc_suffix}.csv"
                )
            )
            t_df.rename_axis("fov", axis=1, inplace=True)
        # If it doesn't exist, transform the data and save it.
        except FileNotFoundError:
            t_df: pd.DataFrame = qc_control.transformed_control_effects_data(
                control_sample_name=control_sample_name, qc_metric=qc_col, to_csv=save_lc_df
            )

        # Set up the Figure for multiple axes
        fig: Figure = plt.figure(figsize=figsize, dpi=dpi)
        fig.set_layout_engine(layout="constrained")
        gs = gridspec.GridSpec(nrows=2, ncols=1, figure=fig, height_ratios=[len(t_df.index) - 1, 1])

        # Colorbar Normalization
        _norm = Normalize(vmin=-1, vmax=1)
        _cmap = sns.color_palette("vlag", as_cmap=True)

        fig.suptitle(f"{control_sample_name} - QC: {qc_col}")

        # Annontation kwargs
        annotation_kws = {
            "horizontalalignment": "center",
            "verticalalignment": "center",
            "fontsize": 8,
        }

        # Heatmap
        ax_heatmap: Axes = fig.add_subplot(gs[0, 0])

        sns.heatmap(
            t_df[~t_df.index.isin(["mean"])],
            ax=ax_heatmap,
            linewidths=1,
            linecolor="black",
            cbar_kws={"shrink": 0.5},
            annot=True,
            annot_kws=annotation_kws,
            xticklabels=False,
            norm=_norm,
            cmap=_cmap,
        )

        # cbar title
        ax_heatmap.collections[0].colorbar.ax.set_title(r"$\log_2(QC)$")

        # Axes labels, and ticks
        ax_heatmap.set_yticks(
            ticks=ax_heatmap.get_yticks(),
            labels=ax_heatmap.get_yticklabels(),
            rotation=0,
        )
        ax_heatmap.set_xlabel(None)

        # Averaged values
        ax_avg: Axes = fig.add_subplot(gs[1, 0])

        sns.heatmap(
            data=t_df[t_df.index.isin(["mean"])],
            ax=ax_avg,
            linewidths=1,
            linecolor="black",
            annot=True,
            annot_kws=annotation_kws,
            fmt=".2f",
            cmap=ListedColormap(["white"]),
            cbar=False,
        )
        ax_avg.set_yticks(
            ticks=ax_avg.get_yticks(),
            labels=["Mean"],
            rotation=0,
        )
        ax_avg.set_xticks(
            ticks=ax_avg.get_xticks(),
            labels=ax_avg.get_xticklabels(),
            rotation=45,
            ha="right",
            rotation_mode="anchor",
        )
        ax_heatmap.set_ylabel("Channel")
        ax_avg.set_xlabel("FOV")

        # Save figure
        if save_figure:
            fig.savefig(
                fname=pathlib.Path(qc_control.metrics_dir)
                / "figures"
                / f"{control_sample_name}_heatmap_{qc_suffix}.png",
                dpi=dpi,
                bbox_inches="tight",
            )
        if display_figure:
            fig.show()
        else:
            plt.close(fig)



from dataclasses import dataclass

from alpineer import io_utils, misc_utils


@dataclass
class VoltageSweepFile:
    """Class containing sweep information."""

    voltage: int
    date: str
    time: str


def parse_sweep_parameters(sweep_name):
    """Performs string manipulations to get the sweep parameters from an FOV folder.

    Args:
        sweep_name: the name of a folder containing a single FOV of a detector sweep

    Returns:
        VoltageSweepFile: the extracted voltage, date, and time
    """
    _, voltage, date, time = sweep_name.split("_")
    voltage = int(voltage[:-1])
    time = time.replace("-", "")

    return VoltageSweepFile(voltage, date, time)


def find_detector_sweep_folders(data_dir, first_fov, last_fov, sweep_step=25):
    """Generates the names of all folders from a single detector sweep.

    Args:
        data_dir: directory where runs are saved
        first_fov: the name of the first FOV from the detector sweep
        last_fov: the name of the last FOV from the detector sweep
        sweep_step: the voltage increase between FOVs in the detector sweep

    Returns:
        list: the names of all detector sweep folders
    """
    first_sweep = parse_sweep_parameters(first_fov)
    last_sweep = parse_sweep_parameters(last_fov)

    if first_sweep.date != last_sweep.date:
        raise ValueError(
            "The day of the first fov {} is not the same asthe day of the last fov {}".format(
                first_sweep.date, last_sweep.date
            )
        )

    # get all folders with matching date
    potential_sweeps = io_utils.list_folders(data_dir, "Detector")
    potential_sweeps = [sweep for sweep in potential_sweeps if first_sweep.date in sweep]

    # only keep folders created between start and end time
    potential_sweeps = [
        sweep
        for sweep in potential_sweeps
        if first_sweep.time < parse_sweep_parameters(sweep).time < last_sweep.time
    ]

    pred_sweep_voltages = range(first_sweep.voltage + sweep_step, last_sweep.voltage, sweep_step)
    obs_sweep_voltages = [parse_sweep_parameters(sweep).voltage for sweep in potential_sweeps]

    misc_utils.verify_same_elements(
        predicted_voltages=pred_sweep_voltages, observed_voltages=obs_sweep_voltages
    )

    final_folder_list = [first_fov] + potential_sweeps + [last_fov]

    return final_folder_list



# tiled regions
REGION_PARAM_FIELDS = [
    "region_name",
    "region_start_row",
    "region_start_col",
    "fov_num_row",
    "fov_num_col",
    "row_fov_size",
    "col_fov_size",
    "region_rand",
]

# mibitracker
MIBITRACKER_BACKEND = "https://backend-dot-mibitracker-angelolab.appspot.com"

# co-registration
COREG_SAVE_PATH = "C:\\Users\\Customer.ION\\Documents\\toffy\\toffy\\coreg_params.json"
FIDUCIAL_POSITIONS = [
    "top left",
    "top right",
    "middle left",
    "middle right",
    "bottom left",
    "bottom right",
]

MICRON_TO_STAGE_X_MULTIPLIER = 0.001001
MICRON_TO_STAGE_X_OFFSET = 0.3116
MICRON_TO_STAGE_Y_MULTIPLIER = 0.001018
MICRON_TO_STAGE_Y_OFFSET = 0.6294

STAGE_LEFT_BOUNDARY = -1.03
STAGE_RIGHT_BOUNDARY = 23.19
STAGE_TOP_BOUNDARY = 56.16
STAGE_BOTTOM_BOUNDARY = 1.57
OPTICAL_LEFT_BOUNDARY = 386.0
OPTICAL_RIGHT_BOUNDARY = 730.1
OPTICAL_TOP_BOUNDARY = 302.1
OPTICAL_BOTTOM_BOUNDARY = 1085.7

OPTICAL_BOUNDARY_TOL = 100
STAGE_BOUNDARY_TOL = 1

COREG_PARAM_BASELINE = {
    "STAGE_TO_OPTICAL_X_MULTIPLIER": 15.05,
    "STAGE_TO_OPTICAL_X_OFFSET": 26.15,
    "STAGE_TO_OPTICAL_Y_MULTIPLIER": -15.03,
    "STAGE_TO_OPTICAL_Y_OFFSET": -76.16,
}


# QC channels to ignore
QC_CHANNEL_IGNORE = ["Au", "Fe", "Na", "Ta", "Noodle"]

# QC metric .csv suffix and column naming
QC_SUFFIXES = ["nonzero_mean_stats", "total_intensity_stats", "percentile_99_9_stats"]
QC_COLUMNS = ["Non-zero mean intensity", "Total intensity", "99.9% intensity value"]



import math
import os
import re
from collections import Counter

import natsort as ns
import numpy as np
import skimage.io as io
import xarray as xr
from alpineer import data_utils, image_utils, io_utils, load_utils, misc_utils
from skimage import transform

from toffy import json_utils


def get_max_img_size(tiff_out_dir, img_sub_folder="", run_dir=None, fov_list=None):
    """Retrieves the maximum FOV image size listed in the run file, or for the given FOVs.

    Args:
        tiff_out_dir (str): path to the extracted images for the specific run
        img_sub_folder (str): optional name of image sub-folder within each fov
        run_dir (str): path to the run directory containing the run json files, default None
        fov_list (list): list of fovs to check max size for, default none which check all fovs
    Returns:
        value of max image size
    """
    if run_dir:
        run_name = os.path.basename(run_dir)
        run_file_path = os.path.join(run_dir, run_name + ".json")

    img_sizes = []

    # check for a run file
    if run_dir and os.path.exists(run_file_path):
        # retrieve all pixel width dimensions of the fovs
        run_data = json_utils.read_json_file(run_file_path)

        if not fov_list:
            for fov in run_data["fovs"]:
                img_sizes.append(fov.get("frameSizePixels")["width"])
        else:
            for fov in fov_list:
                fov_digits = re.findall(r"\d+", fov)
                run = run_data.get("fovs")
                # get data for fov in list
                fov_data = list(
                    filter(
                        lambda fov: fov["runOrder"] == int(fov_digits[0])
                        and fov["scanCount"] == int(fov_digits[1]),
                        run,
                    )
                )
                img_sizes.append(fov_data[0].get("frameSizePixels")["width"])

    # use extracted images to get max size
    else:
        if not fov_list:
            fov_list = io_utils.list_folders(tiff_out_dir, substrs="fov-")
        channels = io_utils.list_files(os.path.join(tiff_out_dir, fov_list[0], img_sub_folder))
        # check image size for each fov
        for fov in fov_list:
            test_file = io.imread(os.path.join(tiff_out_dir, fov, img_sub_folder, channels[0]))
            img_sizes.append(test_file.shape[1])

    # largest in run
    max_img_size = max(img_sizes)
    return max_img_size


def get_tiled_names(fov_list, run_dir):
    """Retrieves the original tiled name for each fov.

    Args:
        fov_list (list): list of fovs that have an existing image dir
        run_dir (str): path to the run directory containing the run json file
    Returns:
        dictionary with RnCm name as keys and the fov-x-scan-1 name as values
    """
    run_name = os.path.basename(run_dir)
    run_file_path = os.path.join(run_dir, run_name + ".json")
    fov_names = {}

    # check for a run file
    io_utils.validate_paths(run_file_path)

    # retrieve all tiled fov names
    run_data = json_utils.read_json_file(run_file_path)
    for fov in run_data["fovs"]:
        run_order = fov.get("runOrder")
        default_name = f"fov-{run_order}-scan-1"

        if default_name in fov_list:
            # get tiled name
            tiled_name = fov.get("name")
            # filter out non-tiled and moly fovs
            if re.search(re.compile(r"(R\+?\d+)(C\+?\d+)"), tiled_name) and tiled_name != "MoQC":
                fov_names[tiled_name] = default_name

    return fov_names


def rescale_stitched_array(img_data, scale):
    """Take an array of stitched image data and rescale the rows and cols dims.

    Args:
        img_data (xarray.DataArray): data array with 4 dimensions (fov, rows, cols, channels)
        scale (int): amount to scale the data up or down
    Returns:
        xarray.DataArray: data reshaped according to scale value while keeping fov/channel info
    """
    # extract dimension info from xarray
    fov_num, rows, cols, chan_num = img_data.shape
    fovs = img_data.fovs.values
    channels = img_data.channels.values

    rescaled_data = rescale_images(np.array(img_data), scale)

    # fill in the metadata
    rescaled_xr = xr.DataArray(
        rescaled_data,
        coords=[fovs, range(int(rows * scale)), range(int(cols * scale)), channels],
        dims=["fovs", "rows", "cols", "channels"],
    )
    return rescaled_xr


def stitch_images(
    tiff_out_dir,
    run_dir=None,
    channels=None,
    img_sub_folder=None,
    tiled=False,
    intensity_scale=200,
    img_size_scale=0.25,
):
    """Creates a new directory containing stitched channel images for the run.

    Args:
        tiff_out_dir (str): path to the extracted images for the specific run
        run_dir (str): path to the run directory containing the run json files, default None
        channels (list): list of channels to produce stitched images for, None will do all
        img_sub_folder (str): optional name of image sub-folder within each fov
        tiled (bool): whether to stitch images back into original tiled shape
        intensity_scale (int): how much to rescale the image values by,
            needed for Photoshop compatibility
        img_size_scale (int/float): amount to scale down image, set to None for no scaling
    """
    io_utils.validate_paths(tiff_out_dir)
    if run_dir:
        io_utils.validate_paths(run_dir)

    # check for previous stitching
    run_name = os.path.basename(tiff_out_dir)
    stitched_dir = os.path.join(tiff_out_dir, f"{run_name}_stitched")
    if tiled:
        stitched_dir = os.path.join(tiff_out_dir, f"{run_name}_tiled")
        if run_dir is None:
            raise ValueError(
                "You must provide the run directory to stitch images into their "
                "original tiled shape."
            )
    if os.path.exists(stitched_dir):
        raise ValueError(f"The stitch_images subdirectory already exists in {tiff_out_dir}")

    folders = io_utils.list_folders(tiff_out_dir, substrs="fov-")
    folders = ns.natsorted(folders)

    # no img_sub_folder, change to empty string to read directly from base folder
    if img_sub_folder is None:
        img_sub_folder = ""

    # retrieve all extracted channel names, or verify the list provided
    if channels is None:
        channels = io_utils.remove_file_extensions(
            io_utils.list_files(
                dir_name=os.path.join(tiff_out_dir, folders[0], img_sub_folder),
                substrs=".tiff",
            )
        )
    else:
        misc_utils.verify_in_list(
            channel_inputs=channels,
            valid_channels=io_utils.remove_file_extensions(
                io_utils.list_files(
                    dir_name=os.path.join(tiff_out_dir, folders[0], img_sub_folder),
                    substrs=".tiff",
                )
            ),
        )

    # get load and stitching args
    tma_folders = None
    if tiled:
        # returns a dict with keys RnCm and values og folder names
        tiled_folders_dict = get_tiled_names(folders, run_dir)
        tma_folders = list(set(folders).difference(set(tiled_folders_dict.values())))
        tma_folders = ns.natsorted(tma_folders)
        try:
            expected_tiles = load_utils.get_tiled_fov_names(
                list(tiled_folders_dict.keys()), return_dims=True
            )
        except AttributeError:
            raise ValueError("FOV names found in the run file were not in tiled (RnCm) format.")

    # make stitched subdir
    os.makedirs(stitched_dir)

    # save the stitched images to the stitched_image subdir
    for chan in channels:
        if tiled:
            for tile in expected_tiles:
                prefix, expected_fovs, num_rows, num_cols = tile
                if prefix == "":
                    prefix = "unnamed_tile"
                # subset the folders_dict for fovs found in the current tile
                tile_dict = {
                    fov: tiled_folders_dict[fov]
                    for fov in expected_fovs
                    if fov in tiled_folders_dict.keys()
                }

                # save to individual tile subdirs
                tile_stitched_dir = os.path.join(stitched_dir, prefix)
                if not os.path.exists(tile_stitched_dir):
                    os.makedirs(tile_stitched_dir)

                image_data = load_utils.load_tiled_img_data(
                    tiff_out_dir,
                    tile_dict,
                    expected_fovs,
                    chan,
                    single_dir=False,
                    img_sub_folder=img_sub_folder,
                )
                if img_size_scale:
                    image_data = rescale_stitched_array(image_data, img_size_scale)

                fname = os.path.join(tile_stitched_dir, chan + "_stitched.tiff")
                stitched = data_utils.stitch_images(image_data, num_cols)
                current_img = stitched.loc["stitched_image", :, :, chan] / intensity_scale
                image_utils.save_image(fname, current_img)

        if tma_folders or not tiled:
            # save to individual tma subdir
            if tma_folders:
                folders = tma_folders
                stitched_subdir = os.path.join(stitched_dir, "TMA")
                if not os.path.exists(stitched_subdir):
                    os.makedirs(stitched_subdir)
            else:
                stitched_subdir = stitched_dir

            num_cols = math.isqrt(len(folders))
            max_img_size = get_max_img_size(tiff_out_dir, img_sub_folder, run_dir)

            image_data = load_utils.load_imgs_from_tree(
                tiff_out_dir,
                img_sub_folder=img_sub_folder,
                fovs=folders,
                channels=[chan],
                max_image_size=max_img_size,
            )
            if img_size_scale:
                image_data = rescale_stitched_array(image_data, img_size_scale)

            fname = os.path.join(stitched_subdir, chan + "_stitched.tiff")
            stitched = data_utils.stitch_images(image_data, num_cols)
            current_img = stitched.loc["stitched_image", :, :, chan] / intensity_scale
            image_utils.save_image(fname, current_img)


def rescale_images(img_data, scale, save_path=None):
    """Rescale image data to a desired shape.

    Args:
        img_data (np.array): data to be reshaped, expected to be either 2 or 4 dimensions
        scale (int): amount to scale the data up or down
        save_path (str): the location to save the tiff file
    Returns:
        numpy.array: data reshaped according to scale value
    """
    # check data dimensions
    if len(img_data.shape) != 2 and len(img_data.shape) != 4:
        raise ValueError("Image data must have either 2 or 4 dimensions.")

    if scale < 1 and (img_data.shape[1] * scale) % 1 != 0:
        raise ValueError("Scale value less than 1 must result in integer dimensions.")

    # preserve fov and channel dimensions if needed
    if len(img_data.shape) == 4:
        scale = (1, scale, scale, 1)

    # rescale dimension data while preserving values
    data_type = img_data.dtype
    rescaled_data = transform.rescale(
        img_data,
        scale,
        mode="constant",
        preserve_range=True,
    )
    rescaled_data = rescaled_data.astype(data_type)

    # overwrite image tiff
    if save_path:
        io_utils.validate_paths(save_path)
        image_utils.save_image(save_path, rescaled_data)

    return rescaled_data


def fix_image_resolutions(resolution_data, extraction_dir):
    """Rescales any images that are a different resolution than the majority in the run.

    Args:
        resolution_data (pd.DataFrame): details the fov names and resolutions
        extraction_dir (str): path to the extracted images dir for the specific run
    """
    io_utils.validate_paths(extraction_dir)

    # only check extracted fovs
    extracted_fovs = io_utils.list_folders(extraction_dir, substrs="fov")
    resolution_data = resolution_data[np.isin(resolution_data.fov, extracted_fovs)]

    # identify majority value, subset data for other resolutions
    resolutions = resolution_data["pixels / 400 microns"]
    correct_res = Counter(resolutions).most_common(1)[0][0]
    res_change = resolution_data[resolution_data["pixels / 400 microns"] != correct_res]

    if len(res_change) == 0:
        print("No resolution scaling needed for any FOVs in this run.")
        return

    # loop through problematic fovs
    for fov in res_change.fov:
        fov_res = res_change[res_change.fov == fov]["pixels / 400 microns"].values
        scale = correct_res / fov_res
        tiff_data = load_utils.load_imgs_from_tree(extraction_dir, fovs=[fov])

        # scale data
        print(f"Changing {fov} from {int(tiff_data.shape[1])} to {int(tiff_data.shape[1]*scale)}.")
        rescaled_data = rescale_images(np.array(tiff_data), float(scale))

        # save every channel image
        for i, channel in enumerate(tiff_data.channels.values):
            channel_data = rescaled_data[0, :, :, i]
            image_utils.save_image(
                os.path.join(extraction_dir, fov, f"{channel}.tiff"), channel_data
            )



import os
import shutil
from collections import Counter

from alpineer import io_utils, misc_utils

from toffy.json_utils import read_json_file, rename_duplicate_fovs, rename_missing_fovs


def merge_partial_runs(cohort_dir, run_string):
    """Combines different runs together into a single folder of FOVs.

    Args:
        cohort_dir (str): the path to the directory containing the run folders
        run_string (str): the substring that each run folder has
    """
    # create folder to hold contents of all partial runs
    output_folder = os.path.join(cohort_dir, run_string)
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # get list of matching subfolders
    # TODO: list_folders does not handle cases such as "run_dup2" correctly
    # need to add a fix in to alpineer to deal with this
    partial_folders = [f for f in os.listdir(cohort_dir) if run_string in f]
    partial_folders = [partial for partial in partial_folders if partial != run_string]

    if len(partial_folders) == 0:
        raise ValueError("No matching folders found for {}".format(run_string))

    # duplicate fovs check
    fovs_by_folder = {
        partial: io_utils.list_folders(os.path.join(cohort_dir, partial))
        for partial in partial_folders
    }
    fov_names = [
        fov for partial_folder_fovs in fovs_by_folder.values() for fov in partial_folder_fovs
    ]
    duplicate_fovs = [fov for fov, count in Counter(fov_names).items() if count > 1]

    # raise error to report any duplicate fovs
    if duplicate_fovs:
        for fov in duplicate_fovs:
            runs = [run_name for run_name, run_fovs in fovs_by_folder.items() if fov in run_fovs]
            print(f"{fov}: {runs}")
        raise ValueError(
            f"The following FOV folders exist in multiple partial runs: {duplicate_fovs}\n"
            "You need to determine which folders to keep and which to delete before merging."
        )

    # loop through each partial run folder and move the fov folders
    for partial in partial_folders:
        fov_folders = io_utils.list_folders(os.path.join(cohort_dir, partial))

        # copy each fov from partial folder into output folder
        for fov in fov_folders:
            new_path = os.path.join(output_folder, fov)
            shutil.move(os.path.join(cohort_dir, partial, fov), new_path)

        # remove partial folder
        shutil.rmtree(os.path.join(cohort_dir, partial))


def combine_runs(cohort_dir):
    """Combines FOVs from different runs together, using the run name as a unique identifier.

    Args:
        cohort_dir (str): path to the directory containing individual run folders
    """
    # get all runs
    run_folders = io_utils.list_folders(cohort_dir)

    # create folder to hold all images
    output_dir = os.path.join(cohort_dir, "image_data")
    os.makedirs(output_dir)

    # loop over each run
    for run in run_folders:
        run_path = os.path.join(cohort_dir, run)

        fovs = io_utils.list_folders(run_path)
        for fov in fovs:
            shutil.move(os.path.join(run_path, fov), os.path.join(output_dir, run + "_" + fov))

        shutil.rmtree(run_path)


def rename_fov_dirs(json_run_path, default_run_dir, output_run_dir=None):
    """Renames FOV directories with default_name to have custom_name sourced from the run JSON file.

    Args:
        json_run_path (str): path to the JSON run file which contains the custom name values
        default_run_dir (str): directory containing default named FOVs
        output_run_dir (str): directory for renamed FOVs. If None, changed in place

    Raises:
        KeyError: issue reading keys from the JSON file
        ValueError: there exist default named directories that are not described in the run file
        UserWarning: not all custom names from the run file have an existing directory

    """
    io_utils.validate_paths(json_run_path)
    io_utils.validate_paths(default_run_dir)

    # check that new_dir doesn't already exist
    if output_run_dir is not None:
        if os.path.exists(output_run_dir):
            raise ValueError(f"The new directory supplied already exists: {output_run_dir}")

    run_metadata = read_json_file(json_run_path)

    # check for missing or duplicate fov names
    run_metadata = rename_missing_fovs(run_metadata)
    run_metadata = rename_duplicate_fovs(run_metadata)

    # retrieve custom names and number of scans for each fov, construct matching default names
    fov_scan = dict()
    for fov in run_metadata.get("fovs", ()):
        custom_name = fov.get("name")
        run_order = fov.get("runOrder")
        scans = fov.get("scanCount")

        # fovs with multiple scans have scan    number specified
        for scan in range(1, scans + 1):
            default_name = f"fov-{run_order}-scan-{scan}"
            name_ext = f"-{scan}" if scans > 1 else ""
            fov_scan[default_name] = custom_name + name_ext

    # retrieve current default directory names, check if already renamed
    old_dirs = io_utils.list_folders(default_run_dir, "fov")
    if len(old_dirs) == 0:
        raise ValueError(f"All FOV folders in {default_run_dir} have already been renamed")

    # check if custom fov names & scan counts match the number of existing default directories
    misc_utils.verify_in_list(
        warn=True, fovs_in_run_file=list(fov_scan.keys()), existing_fov_folders=old_dirs
    )
    misc_utils.verify_in_list(existing_fov_folders=old_dirs, fovs_in_run_file=list(fov_scan.keys()))

    # if no output specified, FOVs will be renamed inplace
    if output_run_dir is None:
        output_run_dir = default_run_dir

    # change the default directory names to custom names
    for folder in old_dirs:
        original_path = os.path.join(default_run_dir, folder)
        new_path = os.path.join(output_run_dir, fov_scan[folder])

        if output_run_dir == default_run_dir:
            # rename in place
            os.rename(original_path, new_path)
        else:
            # copy to new folder
            shutil.copytree(original_path, new_path)


def rename_fovs_in_cohort(run_names, processed_base_dir, cohort_path, bin_base_dir):
    """Renames the FOVs in each of the supplied runs.

    Args:
        run_names (list): list of runs to rename
        processed_base_dir (str): the directory containing the processed data for each
        cohort_path (str): the path to the folder where renamed FOVs will be saved
        bin_base_dir (str): the directory holding the bin files for each run
    """
    for run in run_names:
        print("Renaming FOVs in {}".format(run))
        input_dir = os.path.join(processed_base_dir, run)
        output_dir = os.path.join(cohort_path, run)
        json_path = os.path.join(bin_base_dir, run, run + ".json")
        rename_fov_dirs(
            json_run_path=json_path,
            default_run_dir=input_dir,
            output_run_dir=output_dir,
        )



"""Helper class for making and retrying requests to the MIBItracker.
Copyright (C) 2021 Ionpath, Inc.  All rights reserved.
"""

import datetime
import io
import json

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError
from skimage import io as skio
from urllib3.util.retry import Retry

# The number of retries will be applied to these code and methods only.
MAX_RETRIES = 3

# Unavailable and Bad Gateway; we see this transiently on App Engine.
RETRY_STATUS_CODES = (502, 503)

# POST is added here as compared to the urllib3 defaults.
RETRY_METHOD_WHITELIST = ("HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE", "POST")

# Timeout for MIBItracker requests
SESSION_TIMEOUT = 10

# Timeout for data transfer requests
DATA_TRANSFER_TIMEOUT = 30


class MibiTrackerError(Exception):
    """Raise for exceptions where the response from the MibiTracker API is
    invalid or unexpected.
    """


class MibiRequests:
    """Helper class for making requests to the MIBItracker.

    This is an opinionated way of using ``requests.Session`` with the following
    features:

    - In the case of specified HTTP errors, requests are retried.
    - All responses call ``raise_for_status()``.
    - An instance requests an authorization token upon initialization, and
      includes it in a default header for all future requests. No token
      refresh capabilities are built in, so if the token expires a new
      instance needs to be initialized.

    Args:
        url:
            The string url to the backend of a MIBItracker instance, e.g.
            ``'https://backend-dot-mibitracker-demo.appspot.com'``.
        email:
            The string email address of your MIBItracker account.
        password:
            The string password of your MIBItracker account.
        token:
            A JSON Web Token (JWT) to validate a MIBItracker session.
        refresh:
            Number of seconds since previous refresh to automatically
            refresh token. Defaults to 5 minutes. Set to 0 or None in order
            to not attempt refreshes.
        retries:
            The max number of retries for HTTP status errors. Defaults
            to ``MAX_RETRIES`` which is set to 3.
        retry_methods:
            The HTTP methods to retry. Defaults to
            ``RETRY_METHOD_WHITELIST`` which is the defaults to ``urllib3``'s
            whitelist with the addition of POST.
        retry_codes:
            The HTTP status codes to retry. Defaults to ``(502, 503)``,
            which are associated with transient errors seen on app engine.
        session_timeout:
            Timeout for MIBItracker requests.
            data_transfer_timeout: Timeout for data transfer requests.

    Attributes:
        url:
            The string url to the backend of a MIBItracker instance.
        session:
            A ``StatusCheckedSession`` that includes an authorization
            header and automatically raises for HTTP status errors.

    Raises:
        HTTPError:
            Raised by ``requests.raise_for_status()``, i.e. if a
            response's status code is >= 400.
    """

    def __init__(
        self,
        url,
        email=None,
        password=None,
        token=None,
        refresh=300,  # 5 minutes
        retries=MAX_RETRIES,
        retry_methods=RETRY_METHOD_WHITELIST,
        retry_codes=RETRY_STATUS_CODES,
        session_timeout=SESSION_TIMEOUT,
        data_transfer_timeout=DATA_TRANSFER_TIMEOUT,
    ):
        """Initialize MibiRequests."""
        self.url = url.rstrip("/")  # We add this as part of request params
        self.session = StatusCheckedSession(timeout=session_timeout)
        self._data_transfer_timeout = data_transfer_timeout
        self._refresh_seconds = refresh
        self._last_refresh = datetime.datetime.now()

        # Provide either an email and password, or a token
        # The token will be used in lieu of email and password if provided
        if token is not None:
            self.session.headers.update({"Authorization": "JWT {}".format(token)})
            self.session.options(self.url)
        elif email is not None and password is not None:
            self._auth(url, email, password)
        else:
            raise ValueError("Provide either both an email and password or a token")

        retry = Retry(
            status=retries,
            method_whitelist=retry_methods,
            status_forcelist=retry_codes,
            backoff_factor=0.3,
        )
        # Set this session to use these retry settings for all http[s] requests
        self.session.mount("http://", HTTPAdapter(max_retries=retry))
        self.session.mount("https://", HTTPAdapter(max_retries=retry))

    def _auth(self, url, email, password):
        """Adds an authorization token to the session's default header."""
        response = self.session.post(
            "{}/api-token-auth/".format(url),
            headers={"content-type": "application/json"},
            data=json.dumps({"email": email, "password": password}),
        )
        token = response.json()["token"]
        self.session.headers.update({"Authorization": "JWT {}".format(token)})

    def refresh(self):
        """Refreshes the authorization token stored in the session header.
        Raises HTTP 400 if attempting to refresh an expired token.
        """
        token = self.session.post(  # use the session to avoid recursion
            "{}/api-token-refresh/".format(self.url),
            data=json.dumps({"token": self.session.headers["Authorization"][4:]}),
            headers={"content-type": "application/json"},
        ).json()["token"]
        self.session.headers.update({"Authorization": "JWT {}".format(token)})

    def _check_refresh(self):
        current_time = datetime.datetime.now()
        if (
            self._refresh_seconds
            and (current_time - self._last_refresh).total_seconds() > self._refresh_seconds
        ):
            self._last_refresh = current_time
            self.refresh()

    @staticmethod
    def _prepare_route(route):
        if not route.startswith("/"):
            return "/{}".format(route)
        return route

    def get(self, route, *args, **kwargs):
        """Makes a GET request to the url using the session.

        Args:
            route:
                The route to add to the base url, such as ``'/images/'``
                or ``'/tissues/?organ=tonsil'``.
            *args:
                Passed to ``requests.Session.get``.
            **kwargs:
                Passes to ``requests.Session.get``.

        Returns:
            requests.Session.get:
                The response from ``requests.Session.get``.
        """
        self._check_refresh()
        return self.session.get(
            "{}{}".format(self.url, self._prepare_route(route)), *args, **kwargs
        )

    def post(self, route, *args, **kwargs):
        """Makes a POST request to the url using the session.

        Args:
            route:
                The route to add to the base url, such as ``'/slides/1/'``.
            *args:
                Passed to ``requests.Session.post``.
            **kwargs:
                Passes to ``requests.Session.post``.

        Returns:
            requests.Session.post:
                The response from ``requests.Session.post``.
        """
        self._check_refresh()
        return self.session.post(
            "{}{}".format(self.url, self._prepare_route(route)), *args, **kwargs
        )

    def put(self, route, *args, **kwargs):
        """Makes a PUT request to the url using the session.

        Args:
            route:
                The route to add to the base url, such as ``'/images/1/'``.
            *args:
                Passed to ``requests.Session.put``.
            **kwargs:
                Passes to ``requests.Session.put``.

        Returns:
            requests.Session.put:
                The response from ``requests.Session.put``.
        """
        self._check_refresh()
        return self.session.put(
            "{}{}".format(self.url, self._prepare_route(route)), *args, **kwargs
        )

    def delete(self, route, *args, **kwargs):
        """Makes a DELETE request to the url using the session.

        Args:
            route:
                The route to add to the base url, such as ``'/images/1/'``.
            *args:
                Passed to ``requests.Session.delete``.
            **kwargs:
                Passes to ``requests.Session.delete``.

        Returns:
            requests.Session.delete:
                The response from ``requests.Session.delete``.
        """
        self._check_refresh()
        return self.session.delete(
            "{}{}".format(self.url, self._prepare_route(route)), *args, **kwargs
        )

    def download_file(self, path):
        """Downloads a file from MIBItracker storage.

        Args:
            path:
                The path to the file in storage. This usually can be
                constructed from the run and image folders.

        Returns:
            io.BytesIO:
                An open file object containing the downloaded file's data,
                rewound to the beginning of the file.
        """
        response = self.get("/download/", params={"path": path})
        url = requests.get(response.json()["url"], timeout=self._data_transfer_timeout)
        buf = io.BytesIO()
        buf.write(url.content)
        buf.seek(0)
        return buf

    def search_runs(self, run_name, run_label=None):
        """Searches for runs which have the name and optionally label.

        Args:
            run_name: The name of the run the image belongs to.
            run_label: (optional) The label of the run.

        Returns:
            list:
                A list of JSON data for each run that matches the search.
                If only ``run_name`` is specified, this list could be of any
                length as a run's name is not necessarily unique. If
                ``run_label`` is also specified, this is guaranteed to be
                unique and the returned list could either be of length
                zero or one.
        """
        payload = {"name": run_name}

        if run_label:
            payload["label"] = run_label

        return self.get("/runs/", params=payload).json()

    def get_channel_data(self, image_id, channel_name):
        """Gets a single channel from MIBItracker as a 2D numpy array.

        Args:
            image_id:
                The integer id of an image.
            channel_name:
                The name of the channel to download.

        Returns:
            numpy.ndarray:
                A MxN numpy array of the channel data.
        """
        try:
            response = self.get(f"images/{image_id}/channel_url/", params={"channel": channel_name})
            response.raise_for_status()
        except HTTPError as e:
            if e.response.status_code == 404:
                raise MibiTrackerError(f"Channel '{channel_name}' not found in the image.")
            raise e

        png = requests.get(response.json()["url"])
        buf = io.BytesIO()
        buf.write(png.content)
        buf.seek(0)
        return skio.imread(buf)


class StatusCheckedSession(requests.Session):
    """Raises for HTTP errors and adds any response JSON to the message."""

    def __init__(self, timeout=SESSION_TIMEOUT):
        """Initialize StatusCheckedSession."""
        super(StatusCheckedSession, self).__init__()
        self.timeout = timeout

    @staticmethod
    def _check_status(response):
        """Check web status."""
        try:
            response.raise_for_status()
        except HTTPError as e:
            try:
                response_json = response.json()
            except json.decoder.JSONDecodeError:
                response_json = None
            raise HTTPError(str(e), response_json, response=response)
        return response

    def _set_timeout(self, kwargs):
        """Timeout set up."""
        if "timeout" not in kwargs:
            kwargs.update({"timeout": self.timeout})

    def get(self, *args, **kwargs):
        """Get current time."""
        self._set_timeout(kwargs)
        response = super().get(*args, **kwargs)
        return self._check_status(response)

    def options(self, *args, **kwargs):
        """Get options."""
        self._set_timeout(kwargs)
        response = super().options(*args, **kwargs)
        return self._check_status(response)

    def post(self, *args, **kwargs):
        """Posts."""
        self._set_timeout(kwargs)
        response = super().post(*args, **kwargs)
        return self._check_status(response)

    def put(self, *args, **kwargs):
        """Puts."""
        self._set_timeout(kwargs)
        response = super().put(*args, **kwargs)
        return self._check_status(response)

    def delete(self, *args, **kwargs):
        """Deletes."""
        self._set_timeout(kwargs)
        response = super().delete(*args, **kwargs)
        return self._check_status(response)



import copy
import itertools
import os
import pathlib
import re
import warnings
from dataclasses import dataclass, field
from shutil import rmtree
from typing import Dict, List, Optional, Tuple, Union

import natsort as ns
import numpy as np
import pandas as pd
import xarray as xr
from alpineer import image_utils, io_utils, load_utils, misc_utils
from pandas.core.groupby import DataFrameGroupBy
from requests.exceptions import HTTPError
from scipy import stats
from scipy.ndimage import gaussian_filter
from tqdm.auto import tqdm

from toffy import settings
from toffy.mibitracker_utils import MibiRequests, MibiTrackerError


def create_mibitracker_request_helper(email, password):
    """Create a mibitracker request helper to access a user's MIBItracker info on Ionpath.

    Args:
        email (str):
            The user's MIBItracker email address
        password (str):
            The user's MIBItracker password

    Returns:
        toffy.mibi.mibitracker_utils.MibiRequests:
            A request helper module instance to access a user's MIBItracker info
    """
    try:
        return MibiRequests(settings.MIBITRACKER_BACKEND, email, password)
    except HTTPError:
        print("Invalid MIBItracker email or password provided")


def download_mibitracker_data(
    email,
    password,
    run_name,
    run_label,
    base_dir,
    tiff_dir,
    overwrite_tiff_dir=False,
    img_sub_folder=None,
    fovs=None,
    channels=None,
):
    """Download a specific run's image data off of MIBITracker
    in an `ark` compatible directory structure.

    Args:
        email (str):
            The user's MIBItracker email address
        password (str):
            The user's MIBItracker password
        run_name (str):
            The name of the run (specified on the user's MIBItracker run page)
        run_label (str):
            The label of the run (specified on the user's MIBItracker run page)
        base_dir (str):
            Where to place the created `tiff_dir`
        overwrite_tiff_dir (bool):
            Whether to overwrite the data already in `tiff_dir`
        tiff_dir (str):
            The name of the data directory in `base_dir` to write the run's image data to
        img_sub_folder (str):
            If specified, the subdirectory inside each FOV folder in `data_dir` to place
            the image data into
        fovs (list):
            A list of FOVs to subset over. If `None`, uses all FOVs.
        channels (lsit):
            A list of channels to subset over. If `None`, uses all channels.

    Returns:
        list:
            A list of tuples containing (point name, point id), sorted by point id.
            This defines the run acquisition order needed for plotting the QC graphs.
    """
    # verify that base_dir provided exists
    if not os.path.exists(base_dir):
        raise FileNotFoundError("base_dir %s does not exist" % base_dir)

    # create the MIBItracker request helper
    mr = create_mibitracker_request_helper(email, password)

    # get the run info using the run_name and the run_label
    # NOTE: there will only be one entry in the 'results' key with run_name and run_label specified
    run_info = mr.search_runs(run_name, run_label)

    # if no results are returned, invalid run_name and/or run_label provided
    if len(run_info["results"]) == 0:
        raise ValueError("No data found for run_name %s and run_label %s" % (run_name, run_label))

    # extract the name of the FOVs and their associated internal IDs
    run_fov_names = [img["number"] for img in run_info["results"][0]["imageset"]["images"]]

    # if fovs is None, ensure all of the fovs in run_fov_names_ids are chosen
    if fovs is None:
        fovs = run_fov_names

    # ensure all of the fovs are valid (important if the user explicitly specifies fovs)
    misc_utils.verify_in_list(provided_fovs=fovs, mibitracker_run_fovs=run_fov_names)

    # iterate through each fov and get the longest set of channels
    # this is to prevent getting no channels if the first FOV is a Moly point or is incomplete
    image_data = run_info["results"][0]["imageset"]["images"]
    _, run_channels = max(
        {i: image_data[i]["pngs"] for i in np.arange(len(image_data))}.items(),
        key=lambda x: len(set(x[1])),
    )

    # if channels is None, ensure all of the channels in run_channels are chosen
    if channels is None:
        channels = run_channels

    # ensure all of the channels are valid (important if the user explicitly specifies channels)
    misc_utils.verify_in_list(provided_chans=channels, mibitracker_run_chans=run_channels)

    # if the desired tiff_dir exists, remove it if overwrite_tiff_dir is True
    # otherwise, throw an error
    if os.path.exists(os.path.join(base_dir, tiff_dir)):
        if overwrite_tiff_dir:
            print("Overwriting existing data in tiff_dir %s" % tiff_dir)
            rmtree(os.path.join(base_dir, tiff_dir))
        else:
            raise ValueError("tiff_dir %s already exists in %s" % (tiff_dir, base_dir))

    # make the image directory
    os.mkdir(os.path.join(base_dir, tiff_dir))

    # ensure sub_folder gets set to "" if img_sub_folder is None (for os.path.join convenience)
    if not img_sub_folder:
        img_sub_folder = ""

    # define the run order list to return
    run_order = []

    # iterate over each FOV of the run
    for img in run_info["results"][0]["imageset"]["images"]:
        # if the image fov name is not specified, move on
        if img["number"] not in fovs:
            continue

        print("Creating data for fov %s" % img["number"])

        # make the fov directory
        os.mkdir(os.path.join(base_dir, tiff_dir, img["number"]))

        # make the img_sub_folder inside the fov directory if specified
        if len(img_sub_folder) > 0:
            os.mkdir(os.path.join(base_dir, tiff_dir, img["number"], img_sub_folder))

        # iterate over each provided channel
        for chan in channels:
            # extract the channel data from MIBItracker as a numpy array
            # fail on the whole FOV if the channel is not found (most likely a Moly point)
            try:
                chan_data = mr.get_channel_data(img["id"], chan)
            except MibiTrackerError:
                print(
                    "On FOV %s, failed to download channel %s, moving on to the next FOV. "
                    "If FOV %s is a Moly point, ignore. "
                    "Otherwise, please ensure that channel %s exists on MibiTracker for FOV %s"
                    % (img["number"], chan, img["number"], chan, img["number"])
                )

                # clean the FOV: we will not have a folder for it (in case of Moly point)
                rmtree(os.path.join(base_dir, tiff_dir, img["number"]))

                # do not attempt to download any more channels
                break

            # define the name of the channel file
            chan_file = "%s.tiff" % chan

            # write the data to a .tiff file in the FOV directory structure
            fname: str = os.path.join(base_dir, tiff_dir, img["number"], img_sub_folder, chan_file)
            image_utils.save_image(fname, chan_data)

        # append the run name and run id to the list
        run_order.append((img["number"], img["id"]))

    return run_order


def compute_nonzero_mean_intensity(image_data):
    """Compute the nonzero mean of a specific fov/chan pair.

    Args:
        image_data (numpy.ndarray):
            the image data for a specific fov/chan pair

    Returns:
        float:
            The nonzero mean intensity of the fov/chan pair (`np.nan` if channel contains all 0s)
    """
    # take just the non-zero pixels
    image_data_nonzero = image_data[image_data != 0]

    # take the mean of the non-zero pixels and assign to (fov, channel) in array
    # unless there are no non-zero pixels, in which case default to 0
    if len(image_data_nonzero) > 0:
        nonzero_mean_intensity = image_data_nonzero.mean()
    else:
        nonzero_mean_intensity = 0

    return nonzero_mean_intensity


def compute_total_intensity(image_data):
    """Compute the sum of all pixels of a specific fov/chan pair.

    Args:
        image_data (numpy.ndarray):
            the image data for a specific fov/chan pair

    Returns:
        float:
            The total intensity of the fov/chan pair (`np.nan` if channel contains all 0s)
    """
    return np.sum(image_data)


def compute_99_9_intensity(image_data):
    """Compute the 99.9% pixel intensity value of a specific fov/chan pair.

    Args:
        image_data (numpy.ndarray):
            the image data for a specific fov/chan pair

    Returns:
        float:
            The 99.9% pixel intensity value of a specific fov/chan pair
    """
    return np.percentile(image_data, q=99.9)


def sort_bin_file_fovs(fovs, suffix_ignore=None):
    """Sort a list of fovs in a bin file by fov and scan number.

    fovs (list):
        a list of fovs prefixed with `'fov-m-scan-n'`
    suffix_ignore (str):
        removes this at the end of each fov name, needed if sorting fov-level QC `.csv` files

    Returns:
        list:
            fov name list sorted by ascending fov number, the ascending scan number
    """
    # set suffix_ignore to the empty string if None
    if suffix_ignore is None:
        suffix_ignore = ""

    # TODO: if anyone can do this using a walrus operator I'd appreciate it!
    return sorted(
        fovs,
        key=lambda f: (
            int(f.replace(suffix_ignore, "").split("-")[1]),
            int(f.replace(suffix_ignore, "").split("-")[3]),
        ),
    )


def compute_qc_metrics(
    extracted_imgs_path, fov_name, gaussian_blur=False, blur_factor=1, save_csv=None
):
    """Compute the QC metric matrices for the image data provided.

    Args:
        extracted_imgs_path (str):
            the directory where extracted images are stored
        fov_name (str):
            the name of the FOV to extract from `bin_file_path`, needs to correspond with JSON name
        gaussian_blur (bool):
            whether or not to add Gaussian blurring
        blur_factor (int):
            the sigma (standard deviation) to use for Gaussian blurring
            set to 0 to use raw inputs without Gaussian blurring
            ignored if `gaussian_blur` set to `False`
        save_csv (str):
            path to save csvs of the qc metrics to

    Returns:
        None
    """
    # path validation checks
    if not os.path.exists(extracted_imgs_path):
        raise FileNotFoundError("extracted_imgs_path %s does not exist" % extracted_imgs_path)

    # retrieve the image data from extracted tiff files
    # the image coords should be: ['fov', 'type', 'x', 'y', 'channel']
    image_data = load_utils.load_imgs_from_tree(extracted_imgs_path, fovs=[fov_name])
    image_data = format_img_data(image_data)

    metric_csvs = compute_qc_metrics_direct(image_data, fov_name, gaussian_blur, blur_factor)
    if save_csv:
        for metric_name, data in metric_csvs.items():
            data.to_csv(os.path.join(save_csv, metric_name), index=False)


def compute_qc_metrics_direct(image_data, fov_name, gaussian_blur=False, blur_factor=1):
    """Compute the QC metric matrices for the image data provided.

    Args:
        image_data (xr.DataArray):
            image data in 'extract_bin_files' output format
        fov_name (str):
            the name of the FOV to extract from `bin_file_path`, needs to correspond with JSON name
        gaussian_blur (bool):
            whether or not to add Gaussian blurring
        blur_factor (int):
            the sigma (standard deviation) to use for Gaussian blurring
            set to 0 to use raw inputs without Gaussian blurring
            ignored if `gaussian_blur` set to `False`

    """
    # there's only 1 FOV and 1 type ('pulse'), so subset on that
    image_data = image_data.loc[fov_name, "pulse", :, :, :]

    # define the list of channels to use
    chans = image_data.channel.values

    # define numpy arrays for all the metrics to extract, more efficient indexing than pandas
    blank_arr = np.zeros(image_data.shape[2], dtype="float32")
    nonzero_mean_intensity = copy.deepcopy(blank_arr)
    total_intensity = copy.deepcopy(blank_arr)
    intensity_99_9 = copy.deepcopy(blank_arr)

    # it's faster to loop through the individual channels rather than broadcasting
    for i, chan in enumerate(chans):
        # subset on the channel, cast to float32 to prevent truncation
        image_data_np = image_data.loc[:, :, chan].values.astype(np.float32)

        # STEP 1: gaussian blur (if specified)
        if gaussian_blur:
            image_data_np = gaussian_filter(
                image_data_np, sigma=blur_factor, mode="nearest", truncate=2.0
            )

        # STEP 2: extract non-zero mean intensity
        nonzero_mean_intensity[i] = compute_nonzero_mean_intensity(image_data_np)

        # STEP 3: extract total intensity
        total_intensity[i] = compute_total_intensity(image_data_np)

        # STEP 4: take 99.9% value of the data and assign
        intensity_99_9[i] = compute_99_9_intensity(image_data_np)

    # define the list of numpy arrays for looping
    metric_data = [nonzero_mean_intensity, total_intensity, intensity_99_9]

    metric_csvs = {}

    for ms, md, mc in zip(settings.QC_SUFFIXES, metric_data, settings.QC_COLUMNS):
        # define the dataframe for this metric
        metric_df = pd.DataFrame(columns=["fov", "channel", mc], dtype=object)

        # assign the metric data
        metric_df[mc] = md

        # assign the fov and channel names
        metric_df["fov"] = fov_name
        metric_df["channel"] = chans

        metric_csvs[f"{fov_name}_{ms}.csv"] = metric_df

    return metric_csvs


def combine_qc_metrics(qc_metrics_dir, warn_overwrite=True):
    """Aggregates the QC results of each FOV into one `.csv`.

    Args:
        qc_metrics_dir (str):
            the name of the folder containing the QC metric files
        warn_overwrite (bool):
            whether to warn if existing combined CSV found for each metric
    """
    # path validation check
    if not os.path.exists(qc_metrics_dir):
        raise FileNotFoundError("qc_metrics_dir %s does not exist" % qc_metrics_dir)

    for ms in settings.QC_SUFFIXES:
        # define an aggregated metric DataFrame
        metric_df = pd.DataFrame()

        # list all the files corresponding to this metric
        metric_files = io_utils.list_files(qc_metrics_dir, substrs=ms + ".csv")

        # don't consider any existing combined .csv files, just the fov-level .csv files
        metric_files = [mf for mf in metric_files if "combined" not in mf]

        # sort the files to ensure consistency
        metric_files = sort_bin_file_fovs(metric_files, suffix_ignore="_%s.csv" % ms)

        # iterate over each metric file and append the data to metric_df
        for mf in metric_files:
            metric_df = pd.concat([metric_df, pd.read_csv(os.path.join(qc_metrics_dir, mf))])

        # write the aggregated metric data
        # NOTE: if this combined metric file already exists, it will be overwritten
        if os.path.exists(os.path.join(qc_metrics_dir, "combined_%s.csv" % ms)) and warn_overwrite:
            warnings.warn(
                "Removing previously generated combined %s file in %s" % (ms, qc_metrics_dir)
            )
        metric_df.to_csv(os.path.join(qc_metrics_dir, "combined_%s.csv" % ms), index=False)


def format_img_data(img_data):
    """Formats the image array from load_imgs_from_tree to be same structure as the array returned
    by extract_bin_files. Works for one FOV data at a time.

    Args:
        img_data (str): current image data array as produced by load function
    Returns:
         xarray.DataArray: image data array with shape [fov, type, x, y, channel]
    """
    # add type dimension
    img_data = img_data.assign_coords(type="pulse")
    img_data = img_data.expand_dims("type", 1)

    # edit dimension names
    img_data = img_data.rename({"fovs": "fov", "rows": "x", "cols": "y", "channels": "channel"})

    return img_data


def qc_filtering(qc_metrics: List[str]) -> Tuple[List[str], List[str]]:
    """Filters the QC columns and suffixes based on the user specified QC metrics,
    then sorts the suffixes w.r.t the columns. Refer to `settings.py` for the
    Column and Suffixes available.

    Args:
        qc_metrics (List[str]): A list of QC metrics to use. Options include:

                - `"Non-zero mean intensity"`
                - `"Total intensity"`
                - `"99.9% intensity value"`


    Returns:
        Tuple[List[str], List[str]]: Returns the QC Columns and the QC Suffixes
    """
    # Filter out unused QC columns and suffixes
    if qc_metrics is not None:
        selected_qcs: List[bool] = [qcm in qc_metrics for qcm in settings.QC_COLUMNS]
        qc_cols = list(itertools.compress(settings.QC_COLUMNS, selected_qcs))
        qc_suffixes = list(itertools.compress(settings.QC_SUFFIXES, selected_qcs))
    else:
        qc_cols: List[str] = settings.QC_COLUMNS
        qc_suffixes: List[str] = settings.QC_SUFFIXES

    return qc_cols, qc_suffixes


def _channel_filtering(
    df: pd.DataFrame, channel_include: List[str] = None, channel_exclude: List[str] = None
) -> pd.DataFrame:
    """Filters the DataFrame based on the included and excluded channels. In addition
    the default ignored channels; Au, Fe, Na, Ta, Noodle, are removed.

    Args:
        df (pd.DataFrame): The DataFrame to filter.
        channel_include (List[str], optional): A list of channels to include. Defaults to None.
        channel_exclude (List[str], optional): A list of channels to exclude. Defaults to None.

    Returns:
        pd.DataFrame: The filtered DataFrame.
    """
    if (
        isinstance(channel_include, list)
        and isinstance(channel_exclude, list)
        and not set(channel_exclude).isdisjoint(set(channel_include))
    ):
        raise ValueError("You cannot include and exclude the same channel.")

    # Filter out the default ignored channels
    df = df[~df["channel"].isin(settings.QC_CHANNEL_IGNORE)]

    # Remove the excluded channels
    # If a channel does not exist, it is ignored
    if channel_exclude is not None:
        df: pd.DataFrame = df[~df["channel"].isin(channel_exclude)]

    # Then filter the excluded channels
    # If a channel does not exist, it is ignored
    if channel_include is not None:
        df: pd.DataFrame = df[df["channel"].isin(channel_include)]
    return df


@dataclass
class QCTMA:
    """Computes the QC metrics for a given list of TMAs of interest and saves TMA specific QC files
    in the `qc_tma_metrics_dir` directory.

    Args:
        cohort_path (Union[str,pathlib.Path]): The directory where the extracted images are
            stored.
        qc_tma_metrics_dir (Union[str, pathlib.path]): The directory where to save the QC TMA
            metrics.
        qc_metrics (List[str]): A list of QC metrics to use. Options include:

                - `"Non-zero mean intensity"`
                - `"Total intensity"`
                - `"99.9% intensity value"`

    Attributes:
        qc_cols (List[str]): A list of the QC columns.
        qc_suffixes (List[str]): A list of the QC suffixes, ordered w.r.t `qc_cols`.
        search_term (re.Pattern): The regex pattern to extract n,m from FOV names of the form RnCm.
        tma_avg_zscores (Dict[str, xr.DataArray]): A dictionary containing the average z-scores for
            each TMA for each QC Metric in `qc_metrics`.
    """

    qc_metrics: Optional[List[str]]
    cohort_path: Union[str, pathlib.Path]
    metrics_dir: Union[str, pathlib.Path]

    # Fields initialized after `__post_init__`
    search_term: re.Pattern = field(init=False)
    qc_cols: List[str] = field(init=False)
    qc_suffixes: List[str] = field(init=False)

    # Set by methods
    tma_avg_zscores: Dict[str, xr.DataArray] = field(init=False)

    def __post_init__(self):
        """Initialize QCTMA."""
        # Input validation: Ensure that the paths exist
        io_utils.validate_paths([self.cohort_path, self.metrics_dir])

        self.qc_cols, self.qc_suffixes = qc_filtering(qc_metrics=self.qc_metrics)

        # Create regex pattern for searching RnCm
        self.search_term: re.Pattern = re.compile(r"R\+?(\d+)C\+?(\d+)")

        # Set the tma_avg_zscores to be an empty dictionary
        self.tma_avg_zscores = {}

    def _get_r_c(self, fov_name: pd.Series) -> pd.Series:
        """Extracts the row and column value from a FOV's name containing RnCm.

        Args:
            fov_name (pd.Series): The FOV's name.

        Returns:
            pd.Series: Returns `n` and `m` as a series of integers.
        """
        r, c = map(int, re.search(self.search_term, fov_name).group(1, 2))
        return pd.Series([r, c])

    def _create_r_c_tma_matrix(
        self, group: DataFrameGroupBy, n_cols: int, n_rows: int, qc_col: str
    ) -> pd.Series:
        """Z-scores all FOVS for a given channel and creates a matrix of size `n_rows` by `n_cols`
        as the TMA grid.

        Args:
            group (DataFrameGroupBy): Each group consists of an individual channel, and all of it's
                associated FOVs.
            n_cols (int): The number of columns in the matrix.
            n_rows (int): The number of rows in the matrix.
            qc_col (str): The column to get the the QC data.

        Returns:
            pd.Series[np.ndarray]: Returns the a series containing the z-score matrix.
        """
        rc_array: np.ndarray = np.full(shape=(n_cols, n_rows), fill_value=np.nan)
        rc_array[group["column"] - 1, group["row"] - 1] = stats.zscore(group[qc_col])

        return pd.Series([rc_array])

    def compute_qc_tma_metrics(self, tmas: List[str]):
        """Calculates the QC metrics for a user specified list of TMAs.

        Args:
            tmas (List[str]): The FOVs with the TMA in the folder name to gather.
        """
        with tqdm(
            total=len(tmas), desc="Computing QC TMA Metrics", unit="TMA", leave=True
        ) as tma_pbar:
            for tma in tmas:
                self._compute_qc_tma_metrics(tma=tma)
                tma_pbar.set_postfix(TMA=tma)
                tma_pbar.update(n=1)

    def _compute_qc_tma_metrics(self, tma: str):
        """Computes the FOV QC metrics for all FOVs in a given TMA.
        If the QC metrics have already been computed, then.

        Args:
            tma (str): The TMA to compute the QC metrics for.
        """
        # cannot use `io_utils.list_folders` because it cannot do a partial "exact match"
        # i.e. if we want to match `tma_1_Rn_Cm` but not `tma_10_Rn_Cm`, `io_utils.list_folders`
        # will return both for `tma_1`
        fovs: List[str] = ns.natsorted(
            seq=(p.name for p in pathlib.Path(self.cohort_path).glob(f"{tma}_*"))
        )

        # Compute the QC metrics
        with tqdm(fovs, desc="Computing QC Metrics", unit="FOV", leave=False) as pbar:
            for fov in pbar:
                # Gather the qc tma files for the current fov if they exist
                pre_computed_metrics = filter(
                    lambda f: "combined" not in f,
                    io_utils.list_files(
                        dir_name=self.metrics_dir,
                        substrs=[f"{fov}_{qc_suffix}.csv" for qc_suffix in self.qc_suffixes],
                    ),
                )

                # only compute if any QC files are missing for the current fov
                if len(list(pre_computed_metrics)) != len(self.qc_cols):
                    compute_qc_metrics(
                        extracted_imgs_path=self.cohort_path,
                        fov_name=fov,
                        save_csv=self.metrics_dir,
                    )
                    pbar.set_postfix(FOV=fov, status="Computing")
                else:
                    pbar.set_postfix(FOV=fov, status="Already Computed")

        # Generate the combined metrics for each TMA
        for qc_suffix in self.qc_suffixes:
            metric_files: List[str] = ns.natsorted(
                (
                    io_utils.list_files(
                        dir_name=self.metrics_dir,
                        substrs=[f"{fov}_{qc_suffix}.csv" for fov in fovs],
                    )
                )
            )

            # Define an aggregated metric DataFrame
            combined_metric_tissue_df: pd.DataFrame = pd.concat(
                (pd.read_csv(os.path.join(self.metrics_dir, mf)) for mf in metric_files)
            )

            combined_metric_tissue_df.to_csv(
                os.path.join(self.metrics_dir, f"{tma}_combined_{qc_suffix}.csv"),
                index=False,
            )

    def qc_tma_metrics_zscore(self, tmas: List[str], channel_exclude: List[str] = None):
        """Creates the average zscore for a given TMA across all FOVs and unexcluded channels.
        By default the following channels are excluded: Au, Fe, Na, Ta, Noodle.

        Args:
            tmas (List[str]): The FOVs withmetet the TMA in the folder name to gather.
            channel_exclude (List[str], optional): An optional list of channels to further filter
                out. Defaults to None.
        """
        max_col, max_row = 0, 0
        with tqdm(total=len(tmas), desc="Computing QC TMA Metric Z-scores", unit="TMA") as pbar:
            for tma in tmas:
                self.tma_avg_zscores[tma] = self._compute_qc_tma_metrics_zscore(
                    tma, channel_exclude=channel_exclude
                )
                max_col = max(self.tma_avg_zscores[tma].shape[1], max_col)
                max_row = max(self.tma_avg_zscores[tma].shape[2], max_row)

                pbar.set_postfix(TMA=tma)
                pbar.update()

        # also average z-scores and store
        all_tmas = np.full(
            shape=(len(tmas), len(self.qc_metrics), max_col, max_row), fill_value=np.nan
        )
        for i, tma in enumerate(tmas):
            col, row = self.tma_avg_zscores[tma].shape[1], self.tma_avg_zscores[tma].shape[2]
            all_tmas[i, :, :col, :row] = self.tma_avg_zscores[tma]

        self.tma_avg_zscores["cross_TMA_averages"] = xr.DataArray(
            data=np.stack(np.nanmean(all_tmas, axis=0)),
            coords=[self.qc_cols, np.arange(max_col), np.arange(max_row)],
            dims=["qc_col", "cols", "rows"],
        )

    def _compute_qc_tma_metrics_zscore(
        self,
        tma: str,
        channel_exclude: List[str] = None,
    ) -> xr.DataArray:
        """Creates the average z-score for a given TMA across all FOVs and unexcluded channels.
        By default the following channels are excluded: Au, Fe, Na, Ta, Noodle.

        Args:
            tma (str): The TMA to compute the average z-score for.
            channel_exclude (List[str], optional): An optional list of channels to further filter
                out. Defaults to None.

        Returns:
            xr.DataArray: An xarray DataArray containing the average z-score for each channel across
                a TMA.
        """
        # Sort the loaded combined csv files based on the filtered `qc_suffixes`
        combined_metric_tmas: List[str] = ns.natsorted(
            io_utils.list_files(self.metrics_dir, substrs=f"{tma}_combined"),
            key=lambda tma_mf: (i for i, qc_s in enumerate(self.qc_suffixes) if qc_s in tma_mf),
        )

        zscore_channels_matrix = []
        n_cols: int = None
        n_rows: int = None

        for cmt, qc_col in zip(combined_metric_tmas, self.qc_cols):
            # Open and filter the default ignored channels, along with the user specified channels
            cmt_df: pd.DataFrame = _channel_filtering(
                df=pd.read_csv(os.path.join(self.metrics_dir, cmt)), channel_exclude=channel_exclude
            )

            cmt_df[["column", "row"]] = cmt_df["fov"].apply(self._get_r_c)

            # Get matrix dimensions
            n_cols = cmt_df["column"].max()
            n_rows = cmt_df["row"].max()

            # Z-score all FOVs per channel, and then create the heatmap matrix
            zscore_channel_tmas: pd.DataFrame = cmt_df.groupby(by="channel", sort=True).apply(
                lambda group: self._create_r_c_tma_matrix(group, n_cols, n_rows, qc_col)
            )
            zscore_channel_matrices: np.ndarray = np.array(
                [c_tma[0] for c_tma in zscore_channel_tmas.values],
            )

            avg_zscore = np.mean(zscore_channel_matrices, axis=0)

            zscore_channels_matrix.append(avg_zscore)

        return xr.DataArray(
            data=np.stack(zscore_channels_matrix),
            coords=[self.qc_cols, np.arange(n_cols), np.arange(n_rows)],
            dims=["qc_col", "cols", "rows"],
        )


@dataclass
class QCControlMetrics:
    """Computes QC Metrics for a set of control sample FOVs across various runs, and saves the QC
    files in the `longitudinal_control_metrics_dir`.

    Args:
        cohort_path (Union[str,pathlib.Path]): The directory where the extracted images are
        stored for the control FOVs.
        longitudinal_control_metrics_dir (Union[str, pathlib.Path]): The directory where to save
        the QC Control metrics.
        qc_metrics (List[str]): A list of QC metrics to use. Options include:

                - `"Non-zero mean intensity"`
                - `"Total intensity"`
                - `"99.9% intensity value"`

    Attributes:
        qc_cols (List[str]): A list of the QC columns.
        qc_suffixes (List[str]): A list of the QC suffixes, ordered w.r.t `qc_cols`.
    """

    qc_metrics: Optional[List[str]]
    cohort_path: Union[str, pathlib.Path]
    metrics_dir: Union[str, pathlib.Path]

    # Fields initialized after `__post_init__`
    qc_cols: List[str] = field(init=False)
    qc_suffixes: List[str] = field(init=False)
    longitudinal_control_metrics: Dict[Tuple[str, str], pd.DataFrame] = field(init=False)

    def __post_init__(self):
        """Initialize QCControlMetrics."""
        # Input validation: Ensure that the paths exist
        io_utils.validate_paths([self.cohort_path, self.metrics_dir])

        self.qc_cols, self.qc_suffixes = qc_filtering(qc_metrics=self.qc_metrics)

        self.longitudinal_control_metrics = {}

    def compute_control_qc_metrics(
        self,
        control_sample_name: str,
        fovs: List[str],
        channel_exclude: List[str] = None,
        channel_include: List[str] = None,
    ) -> None:
        """Computes QC metrics for a set of Control Sample FOVs and saves their QC files in the
        `longitudinal_control_metrics_dir`. Calculates the following metrics for the specified
        control samples:
                - `"Non-zero mean intensity"`
                - `"Total intensity"`
                - `"99.9% intensity value"`.

        Args:
            control_sample_name (str): An identifier for naming the control sample.
            fovs (List[str]): A list of control samples to find QC metrics for.
            channel_exclude (List[str], optional): A list of channels to exclude. Defaults to None.
            channel_include (List[str], optional): A list of channels to include. Defaults to None.


        Raises:
            ValueError: Errors if `tissues` is either None, or a list of size 0.
        """
        if fovs is None or not isinstance(fovs, list):
            raise ValueError("The tissues must be specified as a list of strings")

        with tqdm(
            total=len(fovs),
            desc=f"Computing QC Longitudinal Control metrics - {control_sample_name}",
            unit="FOVs",
        ) as pbar:
            for fov in ns.natsorted(fovs):
                # Gather the qc files for the current fov if they exist
                pre_computed_metrics = filter(
                    lambda f: "combined" not in f,
                    io_utils.list_files(
                        dir_name=self.metrics_dir,
                        substrs=[f"{fov}_{qc_suffix}.csv" for qc_suffix in self.qc_suffixes],
                    ),
                )

                if len(list(pre_computed_metrics)) != len(self.qc_cols):
                    compute_qc_metrics(
                        extracted_imgs_path=self.cohort_path,
                        fov_name=fov,
                        save_csv=self.metrics_dir,
                    )
                    pbar.set_postfix(FOV=fov, status="Computing")
                else:
                    pbar.set_postfix(FOV=fov, status="Already Computed")
                pbar.update()

        # Combine metrics for the set of FOVs into a single file per QC metric
        for qc_col, qc_suffix in zip(self.qc_cols, self.qc_suffixes):
            metric_files = filter(
                lambda f: "combined" not in f,
                io_utils.list_files(
                    dir_name=self.metrics_dir,
                    substrs=[f"{fov}_{qc_suffix}.csv" for fov in fovs],
                ),
            )

            # Define an aggregated metric DataFrame, and filter channels
            combined_lc_df: pd.DataFrame = _channel_filtering(
                df=pd.concat(
                    (pd.read_csv(os.path.join(self.metrics_dir, mf)) for mf in metric_files),
                ),
                channel_include=channel_include,
                channel_exclude=channel_exclude,
            )

            self.longitudinal_control_metrics.update(
                {(control_sample_name, qc_col): combined_lc_df}
            )

            combined_lc_df.to_csv(
                os.path.join(self.metrics_dir, f"{control_sample_name}_combined_{qc_suffix}.csv"),
                index=False,
            )

    def transformed_control_effects_data(
        self, control_sample_name: str, qc_metric: str, to_csv: bool = False
    ) -> pd.DataFrame:
        """Creates a transformed DataFrame for the Longitudinal Control effects data, normalizing by the mean,
        then taking the `log2` of each value.

        Args:
            control_sample_name (str): A control sample to tranform the longitudinal control effects for.
            qc_metric (str): The metric to transform.
            to_csv (bool, optional): Whether to save the transformed data to a csv. Defaults to False.

        Returns:
            pd.DataFrame: The transformed QC Longitudinal Control data.
        """
        misc_utils.verify_in_list(user_metric=qc_metric, qc_metrics=self.qc_cols)

        try:
            df: pd.DataFrame = self.longitudinal_control_metrics[control_sample_name, qc_metric]
        except KeyError:
            # A qc file which isn't stored in the longitudinal_control_metrics dictionary, try to load it
            # in if it exists as a file
            df: pd.DataFrame = pd.read_csv(
                os.path.join(
                    self.metrics_dir,
                    f"{control_sample_name}_combined_{self.qc_suffixes[self.qc_cols.index(qc_metric)]}.csv",
                )
            )
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"QC Metric Not Found for the Control Sample {control_sample_name}"
            ) from e

        # Apply a log2 transformation to the mean normalized data.
        log2_norm_df: pd.DataFrame = df.pivot(
            index="channel", columns="fov", values=qc_metric
        ).transform(func=lambda row: np.log2(row / row.mean()), axis=1)

        mean_log2_norm_df: pd.DataFrame = (
            log2_norm_df.mean(axis=0)
            .to_frame(name="mean")
            .transpose()
            .sort_values(by="mean", axis=1)
        )

        transformed_df: pd.DataFrame = pd.concat(
            objs=[log2_norm_df, mean_log2_norm_df]
        ).sort_values(by="mean", axis=1, inplace=False)

        transformed_df.rename_axis("channel", axis=0, inplace=True)
        transformed_df.rename_axis("fov", axis=1, inplace=True)

        # Save the pivoted dataframe to a csv
        if to_csv:
            qc_suffix: str = self.qc_suffixes[self.qc_cols.index(qc_metric)]
            transformed_df.to_csv(
                os.path.join(
                    self.metrics_dir,
                    f"{control_sample_name}_transformed_{qc_suffix}.csv",
                ),
                index=True,
            )

        return transformed_df



import copy
import os
import random
import shutil
import subprocess
import warnings

import natsort as ns
import numpy as np
import pandas as pd
import skimage.io as io
from alpineer import image_utils, io_utils, load_utils, misc_utils
from scipy.ndimage import gaussian_filter

from toffy.image_stitching import rescale_images
from toffy.json_utils import read_json_file
from toffy.streak_detection import streak_correction


def transform_compensation_json(json_path, comp_mat_path):
    """Converts the JSON file from ionpath into a compensation matrix.

    Args:
        json_path (str): path to json file
        comp_mat_path (str): path to comp matrix

    Returns:
        pd.DataTable: matrix with sources channels as rows and target channels as columns
    """
    data = read_json_file(json_path)["Data"]

    comp_mat = pd.read_csv(comp_mat_path, index_col=0)

    for i in range(len(data)):
        current_data = data[i]
        source = current_data["DonorMass"]
        target = current_data["RecipientMass"]
        val = current_data["Percent"] / 100
        comp_mat.loc[source, str(target)] = val

    comp_mat.to_csv(comp_mat_path)


def _compensate_matrix_simple(raw_inputs, comp_coeffs, out_indices):
    """Perform compensation on the raw data using the supplied compensation values.

    Args:
        raw_inputs (numpy.ndarray):
            array with shape [fovs, rows, cols, channels] containing the image data
        comp_coeffs (numpy.ndarray):
            2D array of coefficients with source channels as rows and targets as columns
        out_indices (numpy.ndarray):
            which indices to generate compensated outputs for

    Returns:
        numpy.ndarray: compensated copy of the raw inputs
    """
    outputs = np.copy(raw_inputs)

    # loop over each channel and construct compensation values
    for chan in out_indices:
        chan_coeffs = comp_coeffs[:, chan]

        # convert from 1D to 4D for broadcasting
        chan_coeffs = np.reshape(chan_coeffs, (1, 1, 1, len(chan_coeffs)))

        # broadcast across entire dataset and collapse into single set of values
        chan_vals = raw_inputs * chan_coeffs
        chan_vals = np.sum(chan_vals, axis=-1)

        # subtract compensated values from target channel
        outputs[..., chan] -= chan_vals

    # set negative values to zero
    outputs = np.where(outputs > 0, outputs, 0)

    # subset on specified indices
    outputs = outputs[..., out_indices]

    return outputs


def validate_inputs(
    raw_data_dir,
    comp_mat,
    acquired_masses,
    acquired_targets,
    input_masses,
    output_masses,
    all_masses,
    fovs,
    save_format,
    raw_data_sub_folder,
    batch_size,
    gaus_rad,
):
    """Helper function to validate inputs for compensate_image_data.

    Args:
        raw_data_dir (str): path to raw data
        comp_mat (pd.DataFrame): compensation matrix
        acquired_masses (list): masses in the supplied panel
        acquired_targets (list): targets in the supplied panel
        input_masses (list): masses to use for compensation
        output_masses (list): masses to compensate
        all_masses (list): masses in the compensation matrix
        fovs (list): fovs in the raw_data_dir
        save_format (str): format to save the data
        raw_data_sub_folder (string): sub-folder for raw images
        batch_size (int): number of images to process concurrently
        gaus_rad (int): radius for smoothing
    """
    # make sure panel is in increasing order
    if not np.all(acquired_masses == sorted(acquired_masses)):
        raise ValueError("Masses must be sorted numerically in the panel_info file")

    # make sure channels in comp matrix are same as those in panel csv
    misc_utils.verify_same_elements(acquired_masses=acquired_masses, compensation_masses=all_masses)

    # check first FOV to make sure all channels are present
    test_data = load_utils.load_imgs_from_tree(
        data_dir=raw_data_dir, fovs=fovs[0:1], img_sub_folder=raw_data_sub_folder
    )

    misc_utils.verify_in_list(
        listed_channels=acquired_targets, image_files=test_data.channels.values
    )

    # make sure supplied masses are present
    if input_masses is not None:
        misc_utils.verify_in_list(input_masses=input_masses, compensation_masses=all_masses)

    if output_masses is not None:
        misc_utils.verify_in_list(output_masses=output_masses, compensation_masses=all_masses)

    # make sure compensation matrix has valid values
    if comp_mat.isna().values.any():
        raise ValueError(
            "Compensation matrix must contain a value for every field; check to "
            "make sure there are no missing values"
        )

    # check for valid save_formats
    allowed_formats = ["raw", "rescaled", "both"]
    misc_utils.verify_in_list(save_format=save_format, allowed_formats=allowed_formats)

    if not isinstance(batch_size, int) or batch_size < 1:
        raise ValueError("batch_size parameter must be a positive integer")

    if not isinstance(gaus_rad, int) or gaus_rad < 0:
        raise ValueError("gaus_rad parameter must be a non-negative integer")


def clean_rosetta_test_dir(folder_path):
    """Remove the unnecessary intermediate folders created by rosetta test data computation.

    Args:
        folder_path (str): base dir for testing, image subdirs will be stored here
    """
    # remove any files beginning with ._, needed to ensure external drive hidden files clear
    _ = subprocess.call(["find", folder_path, "-type", "f", "-name", "._*", "-delete"])

    # remove the compensated data folders
    comp_folders = io_utils.list_folders(folder_path, substrs="compensated_data_")
    for cf in comp_folders:
        shutil.rmtree(os.path.join(folder_path, cf))

    # remove the stitched image folder
    shutil.rmtree(os.path.join(folder_path, "stitched_images"))


def combine_compensation_files(comp_matrix_path, compensation_matrix_names, final_matrix_name):
    """Combine a list of round two compensation matrix files in a given cohort folder.
    This is done additively since round two compensation files are mutually exclusive w.r.t.
    output channels.

    Args:
        comp_matrix_path (str):
            Path to the compensation matrix files to combine
        compensation_matrix_names (list):
            List of files inside `cohort_folder_path` to combine
        final_matrix_name (str):
            Where to write the combined compensation matrix to
    """
    # load in the first matrix inside compensation_matrix_names
    final_compensation_matrix = pd.read_csv(
        os.path.join(comp_matrix_path, compensation_matrix_names[0])
    )

    # loop over the rest and add them in
    # NOTE: skip the first column, since that just delineates the channels
    for matrix in compensation_matrix_names[1:]:
        final_compensation_matrix.iloc[:, 1:] = final_compensation_matrix.iloc[:, 1:].add(
            pd.read_csv(os.path.join(comp_matrix_path, matrix)).iloc[:, 1:]
        )

    # save the final compensation matrix to final_matrix_name
    final_compensation_matrix.to_csv(os.path.join(comp_matrix_path, final_matrix_name), index=False)


def flat_field_correction(img, gaus_rad=100):
    """Apply flat field correction to an image.

    Args:
        img (np.ndarray): image to be corrected
        gaus_rad (int): radius for smoothing

    Returns:
        np.ndarray: corrected image
    """
    # if image is empty, return empty image
    if not np.any(img):
        warnings.warn("Image for flatfield correction is empty")
        return img

    # smooth image
    img_smooth = gaussian_filter(img, sigma=gaus_rad)

    # calculate mean to preserve overall intensity
    img_mean = np.mean(img)

    # apply correction
    img_corr = (img / img_smooth) * img_mean

    return img_corr


def get_masses_from_channel_names(names, panel_df):
    """Get the weights for the given names from the input dataframe.

    Args:
        names (list): the channels whose masses will be returned
        panel_df (pd.DataFrame): the panel containing the masses and channel names

    Returns:
        list: the masses for the given channels
    """
    misc_utils.verify_in_list(
        supplied_channel_names=names, panel_channel_names=panel_df["Target"].values
    )

    weights = panel_df.loc[np.isin(panel_df["Target"], names)]["Mass"].values

    return weights


def compensate_image_data(
    raw_data_dir,
    comp_data_dir,
    comp_mat_path,
    panel_info,
    input_masses=None,
    output_masses=None,
    save_format="rescaled",
    raw_data_sub_folder="",
    batch_size=1,
    gaus_rad=1,
    norm_const=200,
    ffc_masses=[39],
    correct_streaks=False,
    streak_chan="Noodle",
):
    """Function to compensate MIBI data with a flow-cytometry style compensation matrix.

    Args:
        raw_data_dir: path to directory containing raw images
        comp_data_dir: path to directory where compensated images will be saved
        comp_mat_path: path to compensation matrix, nxn with channel labels
        panel_info: array with information about the panel
        input_masses (list): masses from the compensation matrix to use for compensation. If None,
            all masses will be used
        output_masses (list): masses from the compensation matrix that will have tifs generated. If
            None, all masses will be used
        save_format: flag to control how the processed tifs are saved. Must be one of:
            'raw': Direct output from the compensation matrix corresponding to number of ion events
                detected per pixel. These will not be viewable in many image processing programs
            'rescaled': all images are divided by 200 to enable visualization. This transform
                has no effect on downstream analysis as it preserves relative expression values
            'both': saves both 'raw' and 'rescaled' images
        raw_data_sub_folder (string): sub-folder for raw images
        batch_size: number of images to process at a time
        gaus_rad: radius for blurring image data. Passing 0 will result in no blurring
        norm_const: constant used for rescaling if save_format == 'rescaled'
        ffc_masses (list): masses that need to be flat field corrected.
        correct_streaks (bool): whether to correct streaks in the image
        streak_chan (str): the channel to use for streak correction
    """
    io_utils.validate_paths([raw_data_dir, comp_data_dir, comp_mat_path])

    # get list of all fovs
    # TODO: list_folders does not handle cases such as "fov0" correctly
    # need to add a fix in to alpineer to deal with this
    fovs = [f for f in os.listdir(raw_data_dir) if "fov" in f]

    # load csv files
    comp_mat = pd.read_csv(comp_mat_path, index_col=0)
    acquired_masses = panel_info["Mass"].values
    acquired_targets = panel_info["Target"].values
    all_masses = comp_mat.columns.values.astype("int")

    # convert ffc mass into ffc channel names
    if ffc_masses is not None:
        ffc_channels = [
            panel_info.loc[panel_info.Mass == mass].Target.values[0] for mass in ffc_masses
        ]
    else:
        ffc_channels = None

    validate_inputs(
        raw_data_dir,
        comp_mat,
        acquired_masses,
        acquired_targets,
        input_masses,
        output_masses,
        all_masses,
        fovs,
        save_format,
        raw_data_sub_folder,
        batch_size,
        gaus_rad,
    )

    # set unused masses to zero
    if input_masses is not None:
        zero_idx = ~np.isin(all_masses, input_masses)
        comp_mat.loc[zero_idx, :] = 0

    if output_masses is None:
        out_indices = np.arange(len(all_masses))
    else:
        out_indices = np.where(np.isin(all_masses, output_masses))[0]

    # loop over each set of FOVs in the batch
    for i in range(0, len(fovs), batch_size):
        print(f"Processing {fovs[i]}")

        # load batch of fovs
        batch_fovs = fovs[i : i + batch_size]
        batch_data = load_utils.load_imgs_from_tree(
            data_dir=raw_data_dir,
            fovs=batch_fovs,
            channels=acquired_targets,
            img_sub_folder=raw_data_sub_folder,
        )

        # convert to float32 for gaussian_filter and rosetta compatibility
        batch_data = batch_data.astype(np.float32)

        # blur data
        if gaus_rad > 0:
            for j in range(batch_data.shape[0]):
                for k in range(batch_data.shape[-1]):
                    batch_data[j, :, :, k] = gaussian_filter(batch_data[j, :, :, k], sigma=gaus_rad)

        if correct_streaks:
            corrected_channels, _ = streak_correction(
                fov_data=batch_data, streak_channel=streak_chan
            )
            batch_data[0] = corrected_channels.values

        # apply flat field correction if specified
        if ffc_channels is not None:
            misc_utils.verify_in_list(
                flat_field_correction_masses=ffc_channels, all_masses=acquired_targets
            )
            for fov in batch_fovs:
                for chan in ffc_channels:
                    raw_img = batch_data.loc[fov, :, :, chan].values
                    batch_data.loc[fov, :, :, chan] = flat_field_correction(raw_img)

        comp_data = _compensate_matrix_simple(
            raw_inputs=batch_data, comp_coeffs=comp_mat.values, out_indices=out_indices
        )

        # save data
        for j in range(batch_data.shape[0]):
            fov_folder = os.path.join(comp_data_dir, batch_data.fovs.values[j])
            os.makedirs(fov_folder)

            # create directories for saving tifs
            if save_format in ["rescaled", "both"]:
                rescale_folder = os.path.join(fov_folder, "rescaled")
                os.makedirs(rescale_folder)

            if save_format in ["raw", "both"]:
                raw_folder = os.path.join(fov_folder, "raw")
                os.makedirs(raw_folder)

            # this may be only a subset of masses, based on output_masses
            for idx, val in enumerate(out_indices):
                channel_name = batch_data.channels.values[val] + ".tiff"

                # save tifs to appropriate directories
                if save_format in ["rescaled", "both"]:
                    save_path = os.path.join(rescale_folder, channel_name)
                    image_utils.save_image(save_path, comp_data[j, :, :, idx] / norm_const)

                if save_format in ["raw", "both"]:
                    save_path = os.path.join(raw_folder, channel_name)
                    image_utils.save_image(save_path, comp_data[j, :, :, idx])


def copy_round_one_compensated_images(
    runs, round_one_comp_folder, round_two_comp_folder, channels_to_copy
):
    """Copies channels that don't need round two compensation to the round two comp folder.

    Args:
        runs (list):
            runs in `round_two_comp_folder` that need copying from `round_one_comp_folder`
        round_one_comp_folder (str):
            path to the round one Rosetta compensated images
        round_two_comp_folder (str):
            path to the round two Rosetta compensated images
        channels_to_copy (list):
            channels to copy from `round_one_comp_folder` to `round_two_comp_folder`
    """
    io_utils.validate_paths([round_one_comp_folder, round_two_comp_folder])

    # verify runs found in both round one and round two Rosetta folders
    r1_runs = io_utils.list_folders(round_one_comp_folder)
    r2_runs = io_utils.list_folders(round_two_comp_folder)
    misc_utils.verify_in_list(provided_runs=runs, round_one_comp_runs=r1_runs)
    misc_utils.verify_in_list(provided_runs=runs, round_two_comp_runs=r2_runs)

    # for each FOV, copy the channel from their r1_runs folder to r2_runs folder
    for run in runs:
        fovs = io_utils.list_folders(os.path.join(round_one_comp_folder, run), substrs="fov")

        for fov in fovs:
            channel_files = io_utils.list_files(
                os.path.join(round_one_comp_folder, run, fov, "rescaled"), substrs=channels_to_copy
            )

            for cf in channel_files:
                shutil.copy(
                    os.path.join(round_one_comp_folder, run, fov, "rescaled", cf),
                    os.path.join(round_two_comp_folder, run, fov, "rescaled", cf),
                )


def create_tiled_comparison(
    input_dir_list,
    output_dir,
    max_img_size,
    img_sub_folder="rescaled",
    channels=None,
    img_size_scale=0.5,
):
    """Creates a tiled image comparing FOVs from all supplied runs for each channel.

    Args:
        input_dir_list: list of directories to compare
        output_dir: directory where tifs will be saved
        max_img_size (int): largest fov image size
        img_sub_folder: subfolder within each input directory to load images from
        channels: list of channels to compare.
        img_size_scale (int/float): amount to scale down image, set to None for no scaling
    """
    test_dir = input_dir_list[0]
    test_fov = io_utils.list_folders(test_dir)[0]
    test_data = load_utils.load_imgs_from_tree(
        data_dir=test_dir,
        fovs=[test_fov],
        img_sub_folder=img_sub_folder,
        channels=channels,
    )

    if not channels:
        channels = test_data.channels.values

    misc_utils.verify_in_list(
        provided_channels=channels, test_data_channels=test_data.channels.values
    )

    channel_num = len(channels)

    # check that all dirs have the same number of fovs and correct subset of channels
    fov_names = io_utils.list_folders(input_dir_list[0])
    for dir_name in input_dir_list[1:]:
        current_folders = io_utils.list_folders(dir_name)
        misc_utils.verify_same_elements(fov_names1=fov_names, fov_names2=current_folders)
        current_channels = load_utils.load_imgs_from_tree(
            data_dir=dir_name, img_sub_folder=img_sub_folder, fovs=current_folders[:1]
        ).channels.values
        misc_utils.verify_in_list(specified_channels=channels, current_channels=current_channels)

    fov_num = len(fov_names)

    # loop over each channel
    for j in range(channel_num):
        # create tiled array of dirs x fovs
        tiled_image = np.zeros(
            (max_img_size * len(input_dir_list), max_img_size * fov_num),
            dtype=test_data.dtype,
        )

        # loop over each fov, and place into columns of tiled array
        for i in range(fov_num):
            start = i * max_img_size
            end = (i + 1) * max_img_size

            # go through each of the directories, read in the images, and place in the right spot
            for idx, key in enumerate(input_dir_list):
                dir_data = load_utils.load_imgs_from_tree(
                    key,
                    channels=channels[j : j + 1],
                    img_sub_folder=img_sub_folder,
                    max_image_size=max_img_size,
                )
                tiled_image[(max_img_size * idx) : (max_img_size * (idx + 1)), start:end] = (
                    dir_data.values[i, :, :, 0]
                )
        # scale image size
        if img_size_scale:
            tiled_image = rescale_images(tiled_image, img_size_scale)
        fname = os.path.join(output_dir, channels[j] + "_comparison.tiff")

        # save, compress to 3 decimal points for space optimization
        image_utils.save_image(fname, np.round(tiled_image, 3))


def add_source_channel_to_tiled_image(
    raw_img_dir,
    tiled_img_dir,
    output_dir,
    source_channel,
    max_img_size,
    img_sub_folder="",
    percent_norm=99,
    img_size_scale=0.5,
):
    """Adds the specified source_channel to the first row of previously generated tiled images.

    Args:
        raw_img_dir (str): path to directory containing the raw images
        tiled_img_dir (str): path to directory contained the tiled images
        output_dir (str): path to directory where outputs will be saved
        img_sub_folder (str): subfolder within raw_img_dir to load images from
        max_img_size (int): largest fov image size
        source_channel (str): the channel which will be prepended to the tiled images
        percent_norm (int): percentile normalization param to enable easy visualization, set to
            None to skip this step
        img_size_scale (int/float): amount to scale down image, set to None for no scaling
    """
    # load source images
    source_imgs = load_utils.load_imgs_from_tree(
        raw_img_dir,
        channels=[source_channel],
        img_sub_folder=img_sub_folder,
        max_image_size=max_img_size,
    )

    # convert stacked images to concatenated row
    source_list = [source_imgs.values[fov, :, :, 0] for fov in range(source_imgs.shape[0])]
    source_row = np.concatenate(source_list, axis=1)
    source_scaled = rescale_images(source_row, img_size_scale)

    # get percentile of source row if percent_norm set, otherwise leave unset
    perc_source = np.percentile(source_scaled, percent_norm) if percent_norm else None

    # normalize by max if percent_norm is 0 for source in case of sparse signal
    if percent_norm and perc_source == 0:
        perc_source = np.percentile(source_scaled, 100)

    # confirm tiled images have expected shape
    tiled_images = io_utils.list_files(tiled_img_dir)
    test_file = io.imread(os.path.join(tiled_img_dir, tiled_images[0]))
    if test_file.shape[1] != source_scaled.shape[1]:
        raise ValueError(
            "Tiled image {} has shape {}, but source image {} hasshape {}".format(
                tiled_images[0], test_file.shape, source_channel, source_scaled.shape
            )
        )

    # loop through each tiled image, prepend source row scaled, and save
    for tile_name in tiled_images:
        current_tile = io.imread(os.path.join(tiled_img_dir, tile_name))

        # if percent_norm set, normalize the source row to be in the same range as the current tile
        # otherwise, just leave as is (divide by 1)
        perc_ratio = 1
        if percent_norm:
            perc_tile = np.percentile(current_tile, percent_norm)

            # normalize by max if percent_norm is 0 for tile in case of sparse signal
            if perc_tile == 0:
                perc_tile = np.percentile(current_tile, 100)

            perc_ratio = perc_source / perc_tile

        rescaled_source = source_scaled / perc_ratio

        # combine together and save, compress to 3 decimal points for space optimization
        combined_tile = np.concatenate([rescaled_source, current_tile])
        save_name = tile_name.split(".tiff")[0] + "_source_" + source_channel + ".tiff"
        image_utils.save_image(os.path.join(output_dir, save_name), np.round(combined_tile, 3))


def replace_with_intensity_image(run_dir, channel="Au", replace=True, fovs=None):
    """Replaces the specified channel with the intensity image of that channel.

    Args:
        run_dir (str): directory containing extracted run data
        channel (str): the channel whose intensity image will be copied over
        fovs (list or None): the subset of fovs within run_dir which will have their
            intensity image copied over. If None, applies to all fovs
        replace (bool): controls whether intensity image is copied over with _intensity appended
            or if it will overwrite existing channel
    """
    all_fovs = io_utils.list_folders(run_dir)

    # ensure supplied folders are valid
    if fovs is not None:
        misc_utils.verify_in_list(specified_folders=fovs, all_folders=all_fovs)
        all_fovs = fovs

    # ensure channel is valid
    test_file = os.path.join(run_dir, all_fovs[0], "intensities", channel + "_intensity.tiff")
    if not os.path.exists(test_file):
        raise FileNotFoundError("Could not find specified file {}".format(test_file))

    # loop through each fov
    for fov in all_fovs:
        # control whether intensity image overwrites previous image or is copied with new name
        if replace:
            suffix = ".tiff"
        else:
            suffix = "_intensity.tiff"
        shutil.copy(
            os.path.join(run_dir, fov, "intensities", channel + "_intensity.tiff"),
            os.path.join(run_dir, fov, channel + suffix),
        )


def remove_sub_dirs(run_dir, sub_dirs, fovs=None):
    """Removes specified sub-folders from fovs in a run.

    Args:
        run_dir (str): path to directory containing fovs
        sub_dirs (list): directories to remove from each fov
        fovs (list): list of fovs to remove dirs from, otherwise removes from all fovs
    """
    all_fovs = io_utils.list_folders(run_dir)

    # ensure supplied folders are valid
    if fovs is not None:
        misc_utils.verify_in_list(specified_folders=fovs, all_folders=all_fovs)
        all_fovs = fovs

    # ensure all sub_dirs exist
    for sub_dir in sub_dirs:
        if not os.path.isdir(os.path.join(run_dir, all_fovs[0], sub_dir)):
            raise ValueError("Did not find {} in {}".format(sub_dir, all_fovs[0]))

    for fov in all_fovs:
        for sub_dir in sub_dirs:
            shutil.rmtree(os.path.join(run_dir, fov, sub_dir))


def create_rosetta_matrices(
    default_matrix,
    save_dir,
    multipliers,
    current_channel_name,
    output_channel_names,
    masses=None,
):
    """Creates a series of compensation matrices for evaluating coefficients.

    Args:
        default_matrix (str): path to the rosetta matrix to use as the default
        save_dir (str): output directory
        multipliers (list): the range of values to multiply the default matrix by
            to get new coefficients
        current_channel_name (str): channel being adjusted
        output_channel_names (list): subset of the channels to compensate for
        masses (list | None): an optional list of masses to include in the multiplication. If
            only a subset of masses are specified, other masses will retain their values
            in all iterations. If None, all masses are included
    """
    # Read input matrix
    comp_matrix = pd.read_csv(default_matrix, index_col=0)
    comp_masses = comp_matrix.index

    # Check that all entries of comp_matrix are numeric
    if not np.issubdtype(comp_matrix.values.dtype, np.number):
        raise ValueError("Compensation matrix must include only numeric entries")

    # Check channel input
    if masses is None:
        masses = comp_masses
    else:
        try:
            masses = [int(x) for x in masses]
        except ValueError:
            raise ValueError("Masses must be provided as integers")
        misc_utils.verify_in_list(specified_masses=masses, rosetta_masses=comp_masses)

    # define the file prefix used for each compensation matrix file
    default_mat_prefix = os.path.basename(default_matrix).split(".csv")[0]
    output_chan_str = "_".join(output_channel_names) if output_channel_names is not None else "all"
    comp_file_prefix = f"{current_channel_name}_{output_chan_str}_{default_mat_prefix}_mult"

    # loop over each specified multiplier and create separate compensation matrix
    for i in multipliers:
        mult_matrix = copy.deepcopy(comp_matrix)

        for j in range(len(comp_matrix)):
            # multiply specified channel by multiplier
            if comp_masses[j] in masses:
                mult_matrix.iloc[j, :] = comp_matrix.iloc[j, :] * i
        comp_name = f"{comp_file_prefix}_{i}.csv"
        mult_matrix.to_csv(os.path.join(save_dir, comp_name))


# TODO: anything with [f for f in os.listdir(...) ...] needs to be changed
# after list_folders with substrs specified is fixed
def copy_image_files(
    cohort_name, run_names, rosetta_testing_dir, extracted_imgs_dir, fovs_per_run=5
):
    """Creates a new directory for rosetta testing and copies over a random subset of
        previously extracted images.

    Args:
        cohort_name (str): name for all combined runs
        run_names (list): gives names of run folders to retrieve extracted images from
        rosetta_testing_dir (str): directory where to create cohort rosetta testing folder
        extracted_imgs_dir (str): directory containing images from each run,
        fovs_per_run: number of fovs from each run to use for testing, default 5

    """
    # path validation
    io_utils.validate_paths([rosetta_testing_dir, extracted_imgs_dir])

    # validate provided run names
    small_runs = []
    for run in run_names:
        if not os.path.exists(os.path.join(extracted_imgs_dir, run)):
            raise ValueError(f"{run} is not a valid run name found in {extracted_imgs_dir}")
        fovs_in_run = [f for f in os.listdir(os.path.join(extracted_imgs_dir, run)) if "fov" in f]
        # check number of fovs in each run
        if len(fovs_in_run) < fovs_per_run:
            small_runs.append(run)

    run_names_process = run_names[:]
    if len(small_runs) > 0:
        # edge case if none of the runs in run_names have length equal to fovs_per_run
        if len(small_runs) == len(run_names_process):
            raise ValueError(
                f"None of the runs specified contain the minimum amount of FOVs ({fovs_per_run}) "
                "defined by the fovs_per_run given"
            )

        # warn user that runs below fovs_per_run threshold will be skipped
        warnings.warn(
            "The following runs will be skipped because they do not contain the minimum amount "
            f"of FOVs ({fovs_per_run}) defined by the fovs_per_run given: {small_runs}."
        )

        run_names_process = list(set(run_names_process).difference(small_runs))

    # make rosetta testing dir and extracted images subdir
    cohort_rosetta_dir = os.path.join(rosetta_testing_dir, cohort_name)
    os.makedirs(os.path.join(cohort_rosetta_dir, "extracted_images"))

    # randomly choose fovs from a run and copy them to the img subdir in rosetta testing dir
    for i, run in enumerate(ns.natsorted(run_names_process)):
        run_path = os.path.join(extracted_imgs_dir, run)

        fovs_in_run = [f for f in os.listdir(run_path) if "fov" in f]
        fovs_in_run = ns.natsorted(fovs_in_run)
        rosetta_fovs = random.sample(fovs_in_run, k=fovs_per_run)

        for fov in rosetta_fovs:
            fov_path = os.path.join(os.path.join(extracted_imgs_dir, run, fov))
            # prepend the run name to each fov
            new_path = os.path.join(
                os.path.join(cohort_rosetta_dir, "extracted_images", run + "_" + fov)
            )
            shutil.copytree(fov_path, new_path)


def rescale_raw_imgs(img_out_dir, scale=200):
    """Rescale image data to be between 0 and 1.

    Args:
        img_out_dir (str): the directory containing extracted images
        scale (int): how much to rescale the image data by

    Returns:
        saves the rescaled images in a subdir
    """
    io_utils.validate_paths(img_out_dir)

    fovs = io_utils.list_folders(img_out_dir)
    for fov in fovs:
        fov_dir = os.path.join(img_out_dir, fov)
        # create subdirectory for the new images
        sub_dir = os.path.join(fov_dir, "rescaled")
        if not os.path.exists(sub_dir):
            os.makedirs(sub_dir)
        chans = io_utils.list_files(fov_dir)
        # rescale each channel image
        for chan in chans:
            fname = os.path.join(sub_dir, chan)
            if not os.path.exists(fname):
                img = io.imread(os.path.join(fov_dir, chan))
                img = (img / scale).astype("float32")
                image_utils.save_image(fname, img)


def generate_rosetta_test_imgs(
    rosetta_mat_path,
    img_out_dir,
    multipliers,
    folder_path,
    panel,
    current_channel_name="Noodle",
    output_channel_names=None,
    gaus_rad=1,
    norm_const=1,
    ffc_masses=[39],
):
    """Compensate example FOV images based on given multipliers.

    Args:
        rosetta_mat_path (str): path to rosetta compensation matrix
        img_out_dir (str): directory where extracted images are stored
        multipliers (list): list of coeffient multipliers to create different matrices for
        folder_path (str): base dir for testing, image subdirs will be stored here
        panel (pd.DataFrame): the panel containing the masses and channel names
        current_channel_name (str): channel being adjusted, default Noodle
        output_channel_names (list): subset of the channels to compensate for, default None is all
        gaus_rad: radius for blurring image data. Passing 0 will result in no blurring
        norm_const: constant used for rescaling
        ffc_masses (list): masses that need to be flat field corrected.

    Returns:
        Create subdirs containing rosetta compensated images for each multiplier and stitched imgs
    """
    io_utils.validate_paths([rosetta_mat_path, img_out_dir, folder_path])

    # get mass information
    current_channel_mass = get_masses_from_channel_names([current_channel_name], panel)

    if output_channel_names is not None:
        output_masses = get_masses_from_channel_names(output_channel_names, panel)
    else:
        output_masses = None

    # generate rosetta matrices for each multiplier
    create_rosetta_matrices(
        default_matrix=rosetta_mat_path,
        save_dir=folder_path,
        multipliers=multipliers,
        current_channel_name=current_channel_name,
        output_channel_names=output_channel_names,
        masses=current_channel_mass,
    )

    # define the file prefix used for each compensation matrix file
    matrix_name = io_utils.remove_file_extensions([os.path.basename(rosetta_mat_path)])[0]
    output_chan_str = "_".join(output_channel_names) if output_channel_names is not None else "all"
    comp_file_prefix = f"{current_channel_name}_{output_chan_str}_{matrix_name}_mult"

    # loop over each multiplier and compensate the data
    rosetta_dirs = [img_out_dir]
    for multiplier in multipliers:
        print(f"Processing images with multiplier {multiplier}")
        rosetta_mat_path = os.path.join(folder_path, f"{comp_file_prefix}_{multiplier}.csv")
        rosetta_out_dir = os.path.join(folder_path, "compensated_data_{}".format(multiplier))
        rosetta_dirs.append(rosetta_out_dir)
        os.makedirs(rosetta_out_dir)
        compensate_image_data(
            raw_data_dir=img_out_dir,
            comp_data_dir=rosetta_out_dir,
            comp_mat_path=rosetta_mat_path,
            raw_data_sub_folder="rescaled",
            panel_info=panel,
            batch_size=1,
            gaus_rad=gaus_rad,
            norm_const=norm_const,
            output_masses=output_masses,
            ffc_masses=ffc_masses,
        )



# adapted from https://stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file
import hashlib
import os
import warnings

from alpineer import io_utils, misc_utils


def get_hash(filepath):
    """Computes the hash of the specified file to verify file integrity.

    Args:
        filepath (str | PathLike): full path to file

    Returns:
        string: the hash of the file
    """
    with open(filepath, "rb") as f:
        file_hash = hashlib.blake2b()
        while chunk := f.read(8192):
            file_hash.update(chunk)
    return file_hash.hexdigest()


def compare_directories(dir_1, dir_2):
    """Compares two directories to ensure all files are present in both with the same hashes.

    Args:
        dir_1: first directory to compare
        dir_2: second directory to compare

    Returns:
        list: a list of files with different hashes between the two directories
    """
    dir_1_folders = io_utils.list_folders(dir_1)
    dir_2_folders = io_utils.list_folders(dir_2)

    if len(dir_1_folders) > 0:
        warnings.warn(
            "The following subfolders were found in the first directory. Sub-folder "
            "contents will not be compared for accuracy, if you want to ensure "
            "successful copying please run this function on those subdirectories. "
            "{}".format(dir_1_folders)
        )

    if len(dir_2_folders) > 0:
        warnings.warn(
            "The following subfolders were found in the second directory. Sub-folder "
            "contents will not be compared for accuracy, if you want to ensure "
            "successful copying please run this function on those subdirectories. "
            "{}".format(dir_2_folders)
        )

    dir_1_files = io_utils.list_files(dir_1)
    dir_2_files = io_utils.list_files(dir_2)

    misc_utils.verify_same_elements(directory_1=dir_1_files, directory_2=dir_2_files)

    bad_files = []
    for file in dir_1_files:
        hash1 = get_hash(os.path.join(dir_1, file))
        hash2 = get_hash(os.path.join(dir_2, file))

        if hash1 != hash2:
            print("Found a file with differing hashes: {}".format(file))
            bad_files.append(file)

    return bad_files



import inspect
import os
import warnings
from dataclasses import dataclass, field
from typing import Iterable

# prevent memory leaking from creating plots that are never shown
import matplotlib
import pandas as pd
import xarray as xr
from alpineer import io_utils, misc_utils
from mibi_bin_tools.bin_files import _write_out, extract_bin_files
from mibi_bin_tools.type_utils import any_true

from toffy.bin_extraction import incomplete_fov_check
from toffy.image_stitching import stitch_images
from toffy.json_utils import missing_fov_check
from toffy.mph_comp import combine_mph_metrics, compute_mph_metrics, visualize_mph
from toffy.normalize import write_mph_per_mass
from toffy.panel_utils import modify_panel_ranges
from toffy.qc_comp import combine_qc_metrics, compute_qc_metrics_direct
from toffy.qc_metrics_plots import visualize_qc_metrics
from toffy.settings import QC_COLUMNS

matplotlib.use("Agg")

RUN_PREREQUISITES = {
    "plot_qc_metrics": set(["generate_qc"]),
    "plot_mph_metrics": set(["generate_mph"]),
    "image_stitching": set(["extract_tiffs"]),
}


# If FovCallbacks ever should pass data to RunCallbacks, make this a dataclass following the
# field structure outlined for __fov_data and __panel in FovCallbacks
@dataclass
class RunCallbacks:
    """Class for run level callbacks in watcher."""

    run_folder: str

    def plot_qc_metrics(self, qc_out_dir: str, warn_overwrite=False, **kwargs):
        """Plots qc metrics generated by the `generate_qc` callback.

        Args:
            qc_out_dir (str):
                Directory containing qc metric csv
            warn_overwrite (bool): whether to warn if existing `_combined.csv` file found,
                needed to curb watcher output if `plot_qc_metrics` set as intermediate callback
            **kwargs (Dict[str, Any]):
                Additional arguments for `toffy.qc_comp.visualize_qc_metrics`.
                Accepted kwargs are

             - axes_size
             - wrap
             - dpi
             - save_dir
        Returns:
            dict:
                Maps each metric name to their respective plot
        """
        # filter kwargs
        valid_kwargs = ["axes_size", "wrap", "dpi", "save_dir"]
        viz_kwargs = {k: v for k, v in kwargs.items() if k in valid_kwargs}
        qc_plots = {}

        combine_qc_metrics(qc_out_dir, warn_overwrite=warn_overwrite)
        for metric_name in QC_COLUMNS:
            qc_plots[metric_name] = visualize_qc_metrics(
                metric_name, qc_out_dir, **viz_kwargs, return_plot=True
            )

        return qc_plots

    def plot_mph_metrics(self, mph_out_dir, plot_dir, warn_overwrite=False, **kwargs):
        """Plots mph metrics generated by the `generate_mph` callback.

        Args:
            mph_out_dir (str): directory containing mph metric csv
            plot_dir (str): director to store the plot to
            warn_overwrite (bool): whether to warn if existing `_combined.csv` file found,
                needed to curb watcher output if `plot_mph_metrics` set as intermediate callback
            **kwargs (Dict[str, Any]):
                Additional arguments for `toffy.mph_comp.visualize_mph`.
                Accepted kwargs are

             - regression
        Returns:
            matplotlib.figure.Figure:
                The figure containing the MPH plot
        """
        if not os.path.exists(plot_dir):
            os.makedirs(plot_dir)

        # filter kwargs
        valid_kwargs = [
            "regression",
        ]
        viz_kwargs = {k: v for k, v in kwargs.items() if k in valid_kwargs}

        # set verbose to false to prevent overwrite error from popping up each FOV
        mph_df = combine_mph_metrics(mph_out_dir, return_data=True, warn_overwrite=warn_overwrite)
        mph_fig = visualize_mph(mph_df, plot_dir, **viz_kwargs, return_plot=True)

        return mph_fig

    def image_stitching(self, tiff_out_dir, **kwargs):
        """Stitches individual FOV channel images together into one tiff.

        Args:
            tiff_out_dir (str): directory containing extracted images
            **kwargs (Dict[str, Any]):
                Additional arguments for `toffy.image_stitching.stitch_images`.
                Accepted kwargs are

             - channels
        """
        # filter kwargs
        valid_kwargs = ["channels"]
        viz_kwargs = {k: v for k, v in kwargs.items() if k in valid_kwargs}

        stitch_images(tiff_out_dir, self.run_folder, **viz_kwargs)

    def check_incomplete_fovs(self, tiff_out_dir, **kwargs):
        """Checks for partial images (even when fully extracted).

        Args:
            tiff_out_dir (str): directory containing extracted images
            **kwargs (Dict[str, Any]):
                Additional arguments for `toffy.bin_extractions.incomplete_fov_check`.
                Accepted kwargs are

             - num_rows
             - num_channels
             - signal_percent
        Raises:
            Warning if any  FOVs have partially generated images
        """
        incomplete_fov_check(self.run_folder, tiff_out_dir)

    def check_missing_fovs(self, **kwargs):
        """Checks for associated bin/json files per FOV.

        Raises:
            Warning if any fov data is missing
            **kwargs (Dict[str, Any]):
                Additional arguments for `toffy.json_utils.missing_fov_check`.
        """
        missing_fov_check(self.run_folder, os.path.basename(self.run_folder))


@dataclass
class FovCallbacks:
    """Class for FOV level callbacks in watcher."""

    run_folder: str
    point_name: str
    overwrite: bool
    __panel: pd.DataFrame = field(default=None, init=False)
    __panel_prof: pd.DataFrame = field(default=None, init=False)
    __fov_data: xr.DataArray = field(default=None, init=False)
    __fov_data_prof: xr.DataArray = field(default=None, init=False)

    def _generate_fov_data(
        self,
        panel: pd.DataFrame,
        extract_prof: bool,
        intensities=["Au", "chan_39"],
        replace=True,
        time_res=0.0005,
        **kwargs,
    ):
        """Extracts data from bin files using the given panel.

        The data and the panel are then cached members of the FovCallbacks object

        Both the deficient and proficient extracted data and panel are computed and cached

        Args:
            panel (pd.DataFrame):
                Panel used for extraction
            extract_prof (bool):
                If set, extract proficient data
            intensities (bool | List[str]):
                Intensities argument for `mibi_bin_tools.bin_files.extract_bin_files`
            replace (bool):
                Whether to replace pulse images with intensity
            time_res (float):
                Time resolution argument for `mibi_bin_tool.bin_files.extract_bin_files`
            **kwargs (dict):
                Unused kwargs for other functions
        """
        self.__fov_data = extract_bin_files(
            data_dir=self.run_folder,
            out_dir=None,
            include_fovs=[self.point_name],
            panel=panel,
            intensities=intensities,
            replace=replace,
            time_res=time_res,
        )
        self.__panel = panel

        if extract_prof:
            # adds an offset of 0.3 to 'Start' and 'Stop' columns, modifying extraction range
            # from (-0.3, 0) to (0, 0.3) for proficient extraction
            panel_prof = modify_panel_ranges(panel, start_offset=0.3, stop_offset=0.3)
            self.__fov_data_prof = extract_bin_files(
                data_dir=self.run_folder,
                out_dir=None,
                include_fovs=[self.point_name],
                panel=panel_prof,
                intensities=intensities,
                replace=replace,
                time_res=time_res,
            )
            self.__panel_prof = panel_prof

    def extract_tiffs(
        self, tiff_out_dir: str, panel: pd.DataFrame, extract_prof: bool = True, **kwargs
    ):
        """Extract tiffs into provided directory, using given panel.

        Done for both the extracted deficient and proficient data

        Args:
            tiff_out_dir (str):
                Path where tiffs are written
            panel (pd.DataFrame):
                Target mass integration ranges
            extract_prof (bool):
                If set, extract mass proficient data
            **kwargs (dict):
                Additional arguments for `mibi_bin_tools.bin_files.extract_bin_files`.
                Accepted kwargs are

             - intensities
             - replace
             - time_res
        """
        if not os.path.exists(tiff_out_dir):
            os.makedirs(tiff_out_dir)

        extracted_img_dir = os.path.join(tiff_out_dir, self.point_name)
        unextracted_chan_tiffs = []

        # in the case all images have been extracted, simply return
        if os.path.exists(extracted_img_dir) and not self.overwrite:
            all_chan_tiffs = [f"{ct}.tiff" for ct in panel["Target"]]
            extracted_chan_tiffs = io_utils.list_files(extracted_img_dir, substrs=".tiff")
            unextracted_chan_tiffs = set(all_chan_tiffs).difference(extracted_chan_tiffs)

            if len(unextracted_chan_tiffs) == 0:
                warnings.warn(f"Images already extracted for FOV {self.point_name}")
                return

        # ensure we don't re-extract channels that have already been extracted
        if unextracted_chan_tiffs and not self.overwrite:
            unextracted_chans = io_utils.remove_file_extensions(unextracted_chan_tiffs)
            panel = panel[panel["Target"].isin(unextracted_chans)]

        if self.__fov_data is None or self.__fov_data_prof is None:
            self._generate_fov_data(panel, extract_prof, **kwargs)

        intensities = kwargs.get("intensities", ["Au", "chan_39"])
        if any_true(intensities) and type(intensities) is not list:
            intensities = list(self.__fov_data.channel.values)

        _write_out(
            img_data=self.__fov_data[0, :, :, :, :].values,
            out_dir=tiff_out_dir,
            fov_name=self.point_name,
            targets=list(self.__fov_data.channel.values),
            intensities=intensities,
        )

        if extract_prof:
            _write_out(
                img_data=self.__fov_data_prof[0, :, :, :, :].values,
                out_dir=tiff_out_dir + "_proficient",
                fov_name=self.point_name,
                targets=list(self.__fov_data.channel.values),
                intensities=intensities,
            )

    def generate_qc(
        self, qc_out_dir: str, panel: pd.DataFrame = None, extract_prof: bool = True, **kwargs
    ):
        """Generates qc metrics from given panel, and saves output to provided directory.

        Args:
            qc_out_dir (str):
                Path where qc_metrics are written
            panel (pd.DataFrame):
                Target mass integration ranges
            extract_prof (bool):
                If set, extract mass proficient data
            **kwargs (dict):
                Additional arguments for `toffy.qc_comp.compute_qc_metrics`. Accepted kwargs are:

             - gaussian_blur
             - blur_factor
        """
        if not os.path.exists(qc_out_dir):
            os.makedirs(qc_out_dir)

        if self.__fov_data is None:
            if panel is None:
                raise ValueError("Must provide panel if fov data is not already generated...")
            self._generate_fov_data(panel, extract_prof, **kwargs)

        qc_metric_paths = [
            os.path.join(qc_out_dir, f"{self.point_name}_nonzero_mean_stats.csv"),
            os.path.join(qc_out_dir, f"{self.point_name}_total_intensity_stats.csv"),
            os.path.join(qc_out_dir, f"{self.point_name}_percentile_99_9_stats.csv"),
        ]
        if all([os.path.exists(qc_file) for qc_file in qc_metric_paths]) and not self.overwrite:
            warnings.warn(f"All QC metrics already extracted for FOV {self.point_name}")
            return

        metric_data = compute_qc_metrics_direct(
            image_data=self.__fov_data,
            fov_name=self.point_name,
            gaussian_blur=kwargs.get("gaussian_blur", False),
            blur_factor=kwargs.get("blur_factor", 1),
        )

        for metric_name, data in metric_data.items():
            data.to_csv(os.path.join(qc_out_dir, metric_name), index=False)

    def generate_mph(self, mph_out_dir, **kwargs):
        """Generates mph metrics from given panel, and saves output to provided directory.

        Args:
            mph_out_dir (str): where to output mph csvs to
            **kwargs (dict):
                Additional arguments for `toffy.mph_comp.compute_mph_metrics`. Accepted kwargs are:

             - mass
             - mass_start
             - mass_stop
        """
        if not os.path.exists(mph_out_dir):
            os.makedirs(mph_out_dir)

        mph_pulse_file = os.path.join(mph_out_dir, f"{self.point_name}-mph_pulse.csv")
        if os.path.exists(mph_pulse_file) and not self.overwrite:
            warnings.warn(f"MPH pulse metrics already extracted for FOV {self.point_name}")
            return

        compute_mph_metrics(
            bin_file_dir=self.run_folder,
            csv_dir=mph_out_dir,
            fov=self.point_name,
            mass=kwargs.get("mass", 98),
            mass_start=kwargs.get("mass_start", 97.5),
            mass_stop=kwargs.get("mass_stop", 98.5),
        )

    def generate_pulse_heights(self, pulse_out_dir: str, panel: pd.DataFrame = None, **kwargs):
        """Generates pulse height csvs from bin files, and saves output to provided directory.

        Args:
            pulse_out_dir (str): where to output pulse height csvs
            panel (pd.DataFrame): Target mass integration ranges
            **kwargs (dict):
                Additional arguments for `toffy.normalize.write_mph_per_mass`. Accepted kwargs are:

             - start_offset
             - stop_offset
        """
        if not os.path.exists(pulse_out_dir):
            os.makedirs(pulse_out_dir)

        pulse_height_file = os.path.join(pulse_out_dir, f"{self.point_name}_pulse_heights.csv")
        if os.path.exists(pulse_height_file) and not self.overwrite:
            warnings.warn(f"Pulse heights per mass already extracted for FOV {self.point_name}")
            return

        write_mph_per_mass(
            base_dir=self.run_folder,
            output_dir=pulse_out_dir,
            fov=self.point_name,
            masses=panel["Mass"].values,
            start_offset=kwargs.get("mass_start", 0.3),
            stop_offset=kwargs.get("mass_stop", 0),
        )


def build_fov_callback(*args, **kwargs):
    """Assembles callbacks to be run for each transferred FoV.

    Args:
        *args (List[str]):
            Names of member functions of `FovCallbacks` to chain together
        **kwargs (Dict[str, Any]):
            Arguments to pass to `FovCallbacks` member functions specified in *args

    Raises:
        ValueError:
            Raised on non-existant member function or missing required kwarg

    Returns:
        Callable[[str, str], None]
            Chained fov callback which will execute all specified callbacks
    """
    # retrieve all 'non-special' methods of FovCallbacks
    methods = [attr for attr in dir(FovCallbacks) if attr[0] != "_"]

    # validate user callback settings
    misc_utils.verify_in_list(arg_strings=args, valid_callbacks=methods)
    for arg in args:
        # check that required (non-keyword) arguments for `arg` is present in passed `**kwargs`
        argnames = inspect.getfullargspec(getattr(FovCallbacks, arg))[0]
        argnames = [argname for argname in argnames if argname != "self"]
        misc_utils.verify_in_list(required_arguments=argnames, passed_arguments=list(kwargs.keys()))

    # construct actual callback
    def fov_callback(run_folder: str, point_name: str, overwrite: bool = False):
        # construct FovCallback object for given FoV
        callback_obj = FovCallbacks(run_folder, point_name, overwrite)

        # for each member, retrieve the member function and run it
        for arg in args:
            if cb := getattr(callback_obj, arg, None):
                cb(**kwargs)
            else:
                # unreachable...
                raise ValueError(f"Could not locate attribute {arg} in FovCallback object")

    return fov_callback


def build_callbacks(
    run_callbacks: Iterable[str],
    intermediate_callbacks: Iterable[str] = None,
    fov_callbacks: Iterable[str] = ("extract_tiffs",),
    **kwargs,
):
    """Deduces and assembles all run & FoV callbacks for the watcher function.

    Args:
        run_callbacks (Iterable[str]):
            List of run callback names.  These will deduce the prerequisite fov callbacks
        intermediate_callbacks (Iterable[str]):
            List of intermediate callback names, these will be subsets of `run_callbacks`
            but overriden to act as `fov_callbacks`
        fov_callbacks (Iterable[str]):
            List of fov callbacks to be run, regardless of prerequisite status
        **kwargs (Dict[str, Any]):
            Arguments to pass to `RunCallbacks` and `FovCallbacks` member functions

    Raises:
        ValueError:
            Raised on non-existant member function or missing required kwarg

    Returns:
        Callable[[None,], None], Callable[[str, str], None]:
            Assembled run callback and fov callback
    """
    methods = [attr for attr in dir(RunCallbacks) if attr[0] != "_"]

    fov_callbacks = set(fov_callbacks)

    misc_utils.verify_in_list(requested_callbacks=run_callbacks, valid_callbacks=methods)
    if intermediate_callbacks:
        misc_utils.verify_in_list(
            intermediate_callbacks=intermediate_callbacks, valid_callbacks=methods
        )

    callbacks_with_prereq = (
        run_callbacks + intermediate_callbacks if intermediate_callbacks else run_callbacks[:]
    )

    for run_cb in callbacks_with_prereq:
        argnames = inspect.getfullargspec(getattr(RunCallbacks, run_cb))[0]
        argnames = [argname for argname in argnames if argname != "self"]

        misc_utils.verify_in_list(required_arguments=argnames, passed_arguments=list(kwargs.keys()))

        fov_callbacks = fov_callbacks.union(RUN_PREREQUISITES.get(run_cb, set()))

    fov_callback = build_fov_callback(*list(fov_callbacks), **kwargs)

    def run_callback(run_folder: str):
        callback_obj = RunCallbacks(run_folder)

        for run_cb in run_callbacks:
            if cb := getattr(callback_obj, run_cb, None):
                cb(**kwargs)
            else:
                # unreachable...
                raise ValueError(f"Could not locate attribute {run_cb} in RunCallbacks object")

    intermediate_callback = None
    if intermediate_callbacks:

        def intermediate_callback(run_folder: str):
            callback_obj = RunCallbacks(run_folder)
            inter_return_vals = {}

            for run_cb in intermediate_callbacks:
                if cb := getattr(callback_obj, run_cb, None):
                    inter_return_vals[cb.__func__.__name__] = cb(**kwargs)
                else:
                    # unreachable...
                    raise ValueError(f"Could not locate attribute {run_cb} in RunCallbacks object")

            return inter_return_vals

    return fov_callback, run_callback, intermediate_callback
